# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

"""contain the SADeltaBetaProcess. Half automatic best delta / beta finder
"""

__authors__ = [
    "H.Payno",
]
__license__ = "MIT"
__date__ = "28/10/2021"


import logging
import os
from copy import copy, deepcopy
from typing import Optional, Union

import h5py
import numpy
from multiprocessing import Pool
from nabu.pipeline.config import get_default_nabu_config
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from tomoscan.esrf.scan.utils import get_data
from tomoscan.io import HDF5File

import tomwer.version
from tomwer.core.process.reconstruction.nabu.nabucommon import (
    ResultsLocalRun,
    ResultSlurmRun,
    ResultsWithStd,
)
from tomwer.core.process.reconstruction.nabu.nabuslices import SingleSliceRunner
from tomwer.core.process.reconstruction.scores import (
    ComputedScore,
    ScoreMethod,
    apply_roi,
    compute_score,
    get_disk_mask_radius,
)
from tomwer.core.utils.locker import FileLockerManager
from tomwer.core.process.task import Task
from tomwer.core.progress import Progress
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils import logconfig
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.io.utils import format_stderr_stdout
from tomwer.core.process.reconstruction.nabu.nabuscores import (
    run_nabu_one_slice_several_config,
)
from tomwer.core.futureobject import FutureTomwerObject


from silx.utils.deprecation import deprecated, deprecated_warning

from ..nabu import utils as nabu_utils
from .params import SADeltaBetaParams

_logger = logging.getLogger(__name__)


DEFAULT_RECONS_FOLDER = "multi_delta_beta_results"


def one_slice_several_db(
    scan: TomwerScanBase,
    configuration: Union[dict, SADeltaBetaParams],
    process_id: Optional[int] = None,
) -> tuple:
    """
    Run a slice reconstruction using nabu per Center Of Rotation (cor) provided
    Then for each compute a score (quality) of the center of rotation

    :param TomwerScanBase scan:
    :param Union[dict,SADeltaBetaParams] configuration:
    :return: cor_reconstructions, outs, errs
             cor_reconstructions is a dictionary of cor as key and a tuple
             (url, score) as value
    :rtype: tuple
    """
    if isinstance(configuration, SADeltaBetaParams):
        configuration = configuration.to_dict()

    task = SADeltaBetaTask(
        process_id=process_id,
        inputs={
            "data": scan,
            "sa_axis_params": configuration,
            "serialize_output_data": False,
        },
    )
    task.run()
    return (
        task.outputs.scores,
        task.outputs.std_out,
        task.outputs.std_err,
        task.outputs.rois,
    )


class SADeltaBetaTask(
    Task,
    SuperviseProcess,
    input_names=("data", "sa_delta_beta_params"),
    output_names=("data", "best_db"),
    optional_input_names=(
        "dry_run",
        "dump_roi",
        "dump_process",
        "serialize_output_data",
        "compute_scores",  # for GUI we want to post pone the score calculation
        "pool_size",
    ),
):
    """
    Main process to launch several reconstruction of a single slice with
    several Center Of Rotation (cor) values
    """

    DEFAULT_POOL_SIZE = 10

    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        SuperviseProcess.__init__(self, process_id=process_id)
        self._dry_run = inputs.get("dry_run", False)
        self._dump_process = inputs.get("dump_process", True)
        self._dump_roi = inputs.get("dump_roi", False)
        self._sa_delta_beta_params = inputs.get("sa_delta_beta_params", None)
        self._std_outs = tuple()
        self._std_errs = tuple()
        self._cancelled = False

    @property
    def dump_roi(self):
        return self._dump_roi

    @dump_roi.setter
    def dump_roi(self, dump):
        self._dump_roi = dump

    @property
    def std_outs(self):
        return self._std_outs

    @property
    def std_errs(self):
        return self._std_errs

    def set_dry_run(self, dry_run):
        self._dry_run = dry_run

    @property
    def dry_run(self):
        return self._dry_run

    @deprecated(replacement="ewoks task inputs.sa_delta_beta_params")
    def set_configuration(self, configuration: dict) -> None:
        if isinstance(configuration, SADeltaBetaParams):
            self._settings = configuration.to_dict()
        elif isinstance(configuration, dict):
            self._settings = configuration
        else:
            raise TypeError(
                "configuration should be an instance of dict or " "SAAxisParams"
            )

    @staticmethod
    def autofocus(scan) -> Optional[float]:
        scores = scan.sa_delta_beta_params.scores
        if scores is None:
            return
        score_method = scan.sa_delta_beta_params.score_method
        best_db, best_score = None, 0
        for cor, (_, score_cls) in scores.items():
            if score_cls is None:  # if score calculation failed
                continue
            score = score_cls.get(score_method)
            if score is None:
                continue
            if score > best_score:
                best_db, best_score = cor, score
        scan.sa_delta_beta_params.autofocus = best_db
        scan.sa_delta_beta_params.value = best_db
        return best_db

    def get_output_dir(self, params: SADeltaBetaParams, scan: TomwerScanBase):
        output_dir = params.output_dir or None

        if params.output_dir is None:
            output_dir = (
                params.nabu_params.get("output", {}).get("location", None) or None
            )
            if output_dir is None:
                output_dir = os.path.join(scan.path, DEFAULT_RECONS_FOLDER)
        return output_dir

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        if scan is None:
            self.outputs.data = scan
            return
        if isinstance(scan, TomwerScanBase):
            scan = scan
        elif isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            raise ValueError(f"input type of {scan}: {type(scan)} is not managed")

        config = copy(self.inputs.sa_delta_beta_params)
        axis = config.get("reconstruction", {}).get("slice_plane", "XY")
        params = SADeltaBetaParams.from_dict(config)

        # insure scan contains some parameter regarding sa delta / beta
        if scan.sa_delta_beta_params is None:
            scan.sa_delta_beta_params = params

        # insure it also contains some axis_params
        if scan.axis_params is None:
            from tomwer.core.process.reconstruction.axis import AxisRP

            scan.axis_params = AxisRP()

        # create dir if does not exists
        params.output_dir = self.get_output_dir(params=params, scan=scan)
        if not os.path.exists(params.output_dir):
            os.makedirs(params.output_dir)

        slice_index = self._preprocess_slice_index(params.slice_indexes)
        delta_beta_s = params.delta_beta_values
        # TODO: check: dry run should only be settable at one location
        dry_run = self._dry_run or params.dry_run
        cluster_config = params.cluster_config
        nabu_config = params.nabu_recons_params

        # step one: complete nabu configuration(s)
        configs = self._config_preprocessing(
            scan=scan,
            config=nabu_config,
            delta_beta_s=delta_beta_s,
            output_dir=params.output_dir,
        )
        # step 2: run reconstructions
        advancement = Progress(
            f"sa-delta-beta - slice {slice_index} of {scan.get_identifier().short_description()}"
        )

        dbs_res = {}
        rois = {}

        try:
            (
                _,
                dbs_res,
                future_tomo_objs,
                self._std_outs,
                self._std_errs,
            ) = self._run_one_slice_several_db(
                scan=scan,
                configs=configs,
                advancement=advancement,
                slice_index=slice_index,
                dry_run=dry_run,
                cluster_config=cluster_config,
                axis=axis,
            )
        except Exception as e:
            _logger.error(e)
            mess = f"sa-delta-beta -nabu- computation for {str(scan)} failed."
            state = DatasetState.FAILED
        else:
            # step 3: wait for future if any
            self._resolve_futures(
                scan=scan,
                nabu_config=next(iter(configs.items()))[
                    1
                ],  # db is not used but paganin and other parameters are. Take the first nabu configuration available
                slice_index=slice_index,
                db_reconstructions=dbs_res,
                future_tomo_objs=future_tomo_objs,
                axis=axis,
            )

            # step 4: run post processing (compute score for each slice)
            if self.get_input_value("compute_scores", True):
                try:
                    dbs_res, rois = self._post_processing(
                        scan=scan,
                        db_reconstructions=dbs_res,
                    )
                except Exception as e:
                    _logger.error(e)
                    mess = f"sa-delta-beta -post-processing- computation for {str(scan)} failed."
                    state = DatasetState.FAILED
                    dbs_res = {}
                else:
                    state = DatasetState.WAIT_USER_VALIDATION
                    self.delta_beta_s = scan.sa_delta_beta_params.autofocus
                    mess = "sa-delta-beta computation succeeded"
            else:
                dbs_res = {}
                state = DatasetState.FAILED
                mess = "couldn't find 'compute_scores'"

        if self._cancelled:
            state = DatasetState.CANCELLED
            mess = "scan cancelled by the user"

        ProcessManager().notify_dataset_state(
            dataset=scan,
            process=self,
            state=state,
            details=self._compute_mess_details(mess),
        )

        scan.sa_delta_beta_params.scores = dbs_res
        best_db = self.autofocus(scan=scan)
        # store nabu recons parameters to be used within the nabu volume for example.

        sc_config = get_default_nabu_config(nabu_fullfield_default_config)
        sc_config.update(nabu_config)
        if best_db is not None:
            sc_config["phase"]["delta_beta"] = (
                best_db,
            )  # warning: at this tage delta_beta expects a list of value
        # store used reconstruction parameters - to be used later on
        scan.nabu_recons_params = sc_config

        # end processing
        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan
        self.outputs.best_db = best_db

        self._process_end(scan=scan, db_res=dbs_res, score_rois=rois)

    def _compute_mess_details(self, mess=""):
        """
        util to join a message and nabu std err and std out
        """
        nabu_logs = []
        for std_err, std_out in zip(self._std_errs, self.std_outs):
            nabu_logs.append(format_stderr_stdout(stdout=std_out, stderr=std_err))
        self._nabu_log = nabu_logs
        nabu_logs.insert(0, mess)
        return "\n".join(nabu_logs)

    def _config_preprocessing(self, scan, config, delta_beta_s, output_dir) -> dict:
        config.get("phase", {}).pop("beam_shape", None)
        # if scan contains some center of position copy it to nabu
        if (
            scan.axis_params is not None
            and scan.axis_params.relative_cor_value is not None
        ):
            if "reconstruction" in config:
                # move the cor value to the nabu reference
                cor_nabu_ref = scan.axis_params.relative_cor_value + scan.dim_1 / 2.0
                config["reconstruction"]["rotation_axis_position"] = str(cor_nabu_ref)

        _logger.info(f"set nabu reconstruction parameters to {scan}")
        scan.nabu_recons_params = config
        res = {}
        for db in delta_beta_s:
            l_config = deepcopy(config)
            if "output" not in config:
                l_config["output"] = {}
            if output_dir is None:
                l_config["output"]["location"] = os.path.join(
                    scan.path, DEFAULT_RECONS_FOLDER
                )
            else:
                l_config["output"]["location"] = output_dir
            # TODO: allow file format modifications
            l_config["output"]["file_format"] = "hdf5"
            if "phase" not in config:
                l_config["phase"] = {}
            l_config["phase"]["delta_beta"] = db
            l_config["phase"]["method"] = "Paganin"
            res[db] = l_config
        return res

    def _run_one_slice_several_db(
        self,
        scan,
        configs,
        slice_index,
        advancement,
        dry_run,
        axis,
        cluster_config: Optional[dict],
    ):
        future_tomo_objs = {}
        success = True
        recons_urls = {}
        std_outs = []
        std_errs = []

        if not isinstance(cluster_config, (dict, type(None))):
            raise TypeError(
                f"'cluster_config' is expected to be a dict or None. Get {type(cluster_config)} instead."
            )

        runners = run_nabu_one_slice_several_config(
            nabu_configs=configs,
            scan=scan,
            slice_index=slice_index,
            dry_run=dry_run,
            file_format="hdf5",
            advancement=advancement,
            cluster_config=cluster_config,
            process_id=self.process_id,
            instanciate_classes_only=True,
            output_file_prefix_pattern=None,
            axis=axis,
        )

        for runner in runners:
            if self._cancelled:
                break

            self._current_processing = runner
            try:
                results = runner.run()
            except TimeoutError as e:
                _logger.error(e)
            else:
                assert isinstance(
                    results, dict
                ), "results should be a dictionary with delta-beta as key and urls as value"

                for db, res in results.items():
                    success = success and res.success
                    if isinstance(res, ResultsWithStd):
                        std_outs.append(res.std_out)
                        std_errs.append(res.std_err)
                    if (
                        isinstance(res, ResultsLocalRun)
                        and len(res.results_identifiers) > 0
                    ):
                        assert (
                            len(res.results_identifiers) == 1
                        ), "only one slice expected"
                        recons_urls[db] = res.results_identifiers[0]
                    if isinstance(res, ResultSlurmRun):
                        future_tomo_obj = FutureTomwerObject(
                            tomo_obj=scan,
                            process_requester_id=self.process_id,
                            futures=res.future_slurm_jobs,
                        )
                        future_tomo_objs[db] = future_tomo_obj

            if advancement is not None:
                advancement.increaseAdvancement(1)

        return success, recons_urls, future_tomo_objs, std_outs, std_errs

    def _post_processing(
        self,
        scan,
        db_reconstructions,
    ):
        post_processing = _PostProcessing(
            scan=scan,
            db_reconstructions=db_reconstructions,
            pool_size=self.get_input_value("pool_size", self.DEFAULT_POOL_SIZE),
        )
        post_processing._cancelled = self._cancelled
        self._current_processing = post_processing
        return post_processing.run()

    def _resolve_futures(
        self,
        scan,
        nabu_config: dict,
        db_reconstructions,
        slice_index,
        axis,
        future_tomo_objs: dict,
    ):
        assert isinstance(nabu_config, dict)
        pag = False
        ctf = False
        if "phase" in nabu_config:
            phase_method = nabu_config["phase"].get("method", "").lower()
            if phase_method in ("pag", "paganin"):
                pag = True
            elif phase_method in ("ctf",):
                ctf = True

        # treat future.
        for db, future_tomo_obj in future_tomo_objs.items():
            if self._cancelled:
                break

            future_tomo_obj.results()
            if future_tomo_obj.cancelled() or future_tomo_obj.exceptions():
                continue
            file_prefix = SingleSliceRunner.get_file_basename_reconstruction(
                scan=scan,
                slice_index=slice_index,
                pag=pag,
                db=int(db) if db is not None else None,
                ctf=ctf,
                axis=axis,
            )
            # retrieve url
            volume_identifier = nabu_utils.get_recons_volume_identifier(
                file_prefix=file_prefix,
                location=nabu_config["output"]["location"],
                file_format=nabu_config.get("file_format", "hdf5"),
                scan=scan,
                slice_index=None,
                axis=axis,
            )

            assert len(volume_identifier) <= 1, "only one slice expected"
            if len(volume_identifier) == 1:
                db_reconstructions[db] = volume_identifier[0]
            else:
                _logger.warning(
                    f"something went wrong with reconstruction of {db} from {str(scan)}"
                )

    @staticmethod
    def _preprocess_slice_index(slice_index):
        if isinstance(slice_index, str):
            if not slice_index == "middle":
                raise ValueError(f"slice index {slice_index} not recognized")
            else:
                return slice_index
        elif not len(slice_index) == 1:
            raise ValueError("only manage one slice")
        else:
            return list(slice_index.values())[0]

    def _process_end(self, scan, db_res, score_rois):
        assert isinstance(scan, TomwerScanBase)
        state = ProcessManager().get_dataset_state(
            dataset_id=scan.get_identifier(), process=self
        )
        if state not in (
            DatasetState.CANCELLED,
            DatasetState.FAILED,
            DatasetState.SKIPPED,
        ):
            try:
                extra = {
                    logconfig.DOC_TITLE: self._scheme_title,
                    logconfig.SCAN_ID: str(scan),
                }
                slice_index = self.inputs.sa_delta_beta_params.get("slice_index", None)

                if db_res is None:
                    info = f"fail to compute delta/beta scores of slice {slice_index} for scan {scan}."
                    _logger.processFailed(info, extra=extra)
                    ProcessManager().notify_dataset_state(
                        dataset=scan,
                        process=self,
                        state=DatasetState.FAILED,
                        details=info,
                    )
                else:
                    info = f"delta/beta scores of slice {slice_index} for scan {scan} computed."
                    _logger.processSucceed(info, extra=extra)
                    ProcessManager().notify_dataset_state(
                        dataset=scan,
                        process=self,
                        state=DatasetState.WAIT_USER_VALIDATION,
                        details=info,
                    )
            except Exception as e:
                _logger.error(e)
            else:
                if self._dump_process:
                    process_idx = SADeltaBetaTask.process_to_tomwer_processes(
                        scan=scan,
                    )
                    if self.dump_roi and process_idx is not None:
                        self.dump_rois(
                            scan, score_rois=score_rois, process_index=process_idx
                        )

    @staticmethod
    def dump_rois(scan, score_rois: dict, process_index: int):
        if score_rois is None or len(score_rois) == 0:
            return
        if not isinstance(score_rois, dict):
            raise TypeError("score_rois is expected to be a dict")
        process_file = scan.process_file
        process_name = f"tomwer_process_{process_index}"

        if scan.saaxis_params.scores in (None, {}):
            return

        def get_process_path():
            return "/".join((scan.entry or "entry", process_name))

        # save it to the file
        with FileLockerManager().get_lock(process_file):
            # needs an extra lock for multiprocessing

            with HDF5File(process_file, mode="a") as h5f:
                nx_process = h5f.require_group(get_process_path())
                score_roi_grp = nx_process.require_group("score_roi")
                for db, roi in score_rois.items():
                    score_roi_grp[str(db)] = roi
                    score_roi_grp[str(db)].attrs["interpretation"] = "image"

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return "semi-automatic delta/beta finder"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        return "Semi automatic center of rotation / axis calculation"

    @staticmethod
    def process_to_tomwer_processes(scan):
        if scan.process_file is not None:
            entry = "entry"
            if isinstance(scan, NXtomoScan):
                entry = scan.entry

            db = None
            if scan.sa_delta_beta_params is not None:
                db = scan.sa_delta_beta_params.selected_delta_beta_value

            process_index = scan.pop_process_index()
            try:
                with scan.acquire_process_file_lock():
                    Task._register_process(
                        process_file=scan.process_file,
                        entry=entry,
                        results={"delta_beta": db if db is not None else "-"},
                        configuration=scan.sa_delta_beta_params.to_dict(),
                        process_index=process_index,
                        overwrite=True,
                        process=SADeltaBetaTask,
                    )
                    SADeltaBetaTask._extends_results(
                        scan=scan, entry=entry, process_index=process_index
                    )
            except Exception as e:
                _logger.warning(
                    f"Fail to register process of with index {process_index}. Reason is {e}"
                )

            return process_index

    @staticmethod
    def _extends_results(scan, entry, process_index):
        process_file = scan.process_file
        process_name = "tomwer_process_" + str(process_index)

        if scan.sa_delta_beta_params.scores in (None, {}):
            return

        def get_process_path():
            return "/".join((entry or "entry", process_name))

        # save it to the file
        with FileLockerManager().get_lock(process_file):
            # needs an extra lock for multiprocessing

            with HDF5File(process_file, mode="a") as h5f:
                nx_process = h5f.require_group(get_process_path())
                if "NX_class" not in nx_process.attrs:
                    nx_process.attrs["NX_class"] = "NXprocess"

                results = nx_process.require_group("results")
                for cor, (url, score) in scan.sa_delta_beta_params.scores.items():
                    results_db = results.require_group(str(cor))
                    for method in ScoreMethod:
                        if method is ScoreMethod.TOMO_CONSISTENCY:
                            continue
                        results_db[method.value] = score.get(method)

                    link_path = os.path.relpath(
                        url.file_path(),
                        os.path.dirname(process_file),
                    )
                    results_db["reconstructed_slice"] = h5py.ExternalLink(
                        link_path, url.data_path()
                    )

    def cancel(self):
        """
        stop current processing
        """
        if self._current_processing is not None:
            self._cancelled = True
            self._current_processing.cancel()


class _PostProcessing:
    """class used to run SA-delta-beta post-processing on reconstructed slices"""

    DEFAULT_POOL_SIZE = 10

    def __init__(self, db_reconstructions, scan, pool_size) -> None:
        self._db_reconstructions = db_reconstructions
        self._scan = scan
        self._cancelled = False
        self.pool_size = pool_size

    @staticmethod
    def compute_score(item: tuple):
        db, (url, data), mask_disk_radius, cancelled = item
        if data is None:
            score = None
            data_roi = None
        else:
            assert data.ndim == 2
            data_roi = apply_roi(data=data, radius=mask_disk_radius, url=url)

            # move data_roi to [0-1] range
            #  preprocessing: get percentile 0 and 99 from image and
            #  "clean" highest and lowest pixels from it
            min_p, max_p = numpy.percentile(data_roi, (1, 99))
            data_roi_int = data_roi[...]
            data_roi_int[data_roi_int < min_p] = min_p
            data_roi_int[data_roi_int > max_p] = max_p
            data_roi_int = (data_roi_int - min_p) / (max_p - min_p)

            score = ComputedScore(
                tv=compute_score(data=data_roi_int, method=ScoreMethod.TV),
                std=compute_score(data=data_roi_int, method=ScoreMethod.STD),
            )
        return {db: (url, score)}, {db: data_roi}

    def run(self):
        datasets = self.load_datasets()
        assert isinstance(datasets, dict)
        mask_disk_radius = get_disk_mask_radius(datasets)
        with Pool(self.pool_size) as pool:
            res = pool.map(
                self.compute_score,
                [
                    (
                        *item,
                        mask_disk_radius,
                        self._cancelled,
                    )
                    for item in datasets.items()
                ],
            )
        scores = {}
        rois = {}
        for mydict in res:
            myscores, myrois = mydict
            scores.update(myscores)
            rois.update(myrois)
        return scores, rois

    @staticmethod
    def _load_dataset(item: tuple):
        db, volume_identifier = item
        slice_url = None
        # in case the try processing fails
        try:
            volume = VolumeFactory.create_tomo_object_from_identifier(volume_identifier)
            volumes_urls = tuple(volume.browse_data_urls())
            if len(volumes_urls) > 1:
                _logger.warning(
                    f"found a volume with mode that one url ({volumes_urls})"
                )
            slice_url = volumes_urls[0]
            data = get_data(slice_url)
        except Exception as e:
            _logger.error(
                f"Fail to compute a score for {volume_identifier}. Reason is {e}"
            )
            return {db: (slice_url, None)}
        else:
            if data.ndim == 3:
                if data.shape[0] == 1:
                    data = data.reshape(data.shape[1], data.shape[2])
                elif data.shape[2] == 1:
                    data = data.reshape(data.shape[0], data.shape[1])
                else:
                    raise ValueError(f"Data is expected to be 2D. Not {data.ndim}D")
            elif data.ndim == 2:
                pass
            else:
                raise ValueError(f"Data is expected to be 2D. Not {data.ndim}D")

            return {db: (slice_url, data)}

    def load_datasets(self):
        with Pool(self.pool_size) as pool:
            res = pool.map(
                self._load_dataset,
                self._db_reconstructions.items(),
            )
        datasets_ = {}
        for mydict in res:
            datasets_.update(mydict)

        return datasets_

    def cancel(self):
        self._cancelled = True


class SADeltaBetaProcess(SADeltaBetaTask):
    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        deprecated_warning(
            name="tomwer.core.process.reconstruction.sadeltabeta.sadeltabeta.SADeltaBetaProcess",
            type_="class",
            reason="improve readibility",
            since_version="1.2",
            replacement="SADeltaBetaTask",
        )
        super().__init__(process_id, varinfo, inputs, node_id, node_attrs, execinfo)



from orangewidget.utils.signals import PartialSummary, summarize

from tomwer.core.cluster import SlurmClusterConfiguration
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.scanbase import TomwerScanBase


@summarize.register(object)  # noqa F811
def summarize_(Object: object):  # noqa F811
    return PartialSummary("any object", "an oject of any type")


@summarize.register(dict)  # noqa F811
def summarize_(configuration: dict):  # noqa F811
    return PartialSummary(
        "any configuration", "any configuration that can be provided to a process"
    )


@summarize.register(SlurmClusterConfiguration)  # noqa F811
def summarize_(cluster_config: SlurmClusterConfiguration):  # noqa F811
    return PartialSummary(
        "cluster configuration",
        "cluster configuration to launch some remote processing",
    )


@summarize.register(TomwerScanBase)  # noqa F811
def summarize_(data: TomwerScanBase):  # noqa F811
    return PartialSummary(
        "dataset with processing history",
        "core object used to ship dataset and history of processing done on this dataset",
    )


@summarize.register(FutureTomwerObject)  # noqa F811
def summarize_(future_data: FutureTomwerObject):  # noqa F811
    return PartialSummary(
        "dataset with pending processing",
        "object used when there is some pending processing (asyncio.future). Can be convert back to `data`",
    )


@summarize.register(BlissScan)  # noqa F811
def summarize_(bliss_scan: BlissScan):  # noqa F811
    return PartialSummary(
        "raw dataset from bliss",
        "object used when debug some processing relative to bliss",
    )



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
contains gui relative to semi-automatic axis calculation
"""


__authors__ = [
    "H. Payno",
]
__license__ = "MIT"
__date__ = "26/02/2021"


import functools
import logging
import weakref
from typing import Union

import numpy
from silx.gui import icons as silxicons
from silx.gui import qt
from silx.gui.dialog.ImageFileDialog import ImageFileDialog
from silx.gui.plot import items
from silx.utils.enum import Enum as _Enum
from tomoscan.esrf.scan.utils import get_data

from tomwer.core.process.reconstruction.saaxis.saaxis import ReconstructionMode
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui import icons
from tomwer.gui.reconstruction.saaxis.sliceselector import SliceSelector
from tomwer.gui.utils.lineselector import QSliceSelectorDialog
from tomwer.gui.utils.slider import LogSlider
from tomwer.gui.visualization.sinogramviewer import SinogramViewer as _SinogramViewer

_logger = logging.getLogger(__name__)


class DetectionAccuracy(_Enum):
    """The different accuracy modes"""

    COARSE = "Coarse"
    FINE = "Fine"
    ACCURATE = "Accurate"
    DEFINED_RANGE = "Defined range"


ACCURACIES = {
    DetectionAccuracy.COARSE: 500,
    DetectionAccuracy.FINE: 10,
    DetectionAccuracy.ACCURATE: 2,
}


class SinogramViewer(_SinogramViewer):
    """Redefine the main SinogramViewer for saaxis purpose:
    * options orientation is horizontal
    * add the display of the cor roi to be displayed
    * rename ApplyButton to load and change the icon
    * hide the line selection (control by the slice selector
    * move the dock widget to the bottom
    * rename subsampling "loaded sinogram subsampling" to provide more context
    """

    sigCorEdited = qt.Signal(float)
    """Signal emitted when the cor position change. Provided in absolute
    position"""
    sigWidthEdited = qt.Signal(float)
    """Signal emitted when the detection accuracy (width) change."""

    def __init__(self, parent):
        super().__init__(parent, opts_orientation=qt.Qt.Horizontal)
        self._fromMarker = None
        self._toMarker = None
        self._corMarker = None
        self._min = None
        self._max = None
        self._estimated_cor = None
        self._other_cors = None
        # hide line selection
        self._options.setLineSelectionVisible(False)
        # change ApplyButton name and icon
        self._loadButton = self._options._buttons.button(qt.QDialogButtonBox.Apply)
        self._loadButton.setText("load")
        style = qt.QApplication.style()
        self._loadButton.setIcon(style.standardIcon(qt.QStyle.SP_BrowserReload))
        # change subsampling name
        self._options._subsamplingLabel.setText("Sinogram subsampling")
        self._options._subsamplingLabel.setToolTip(
            "Define subsampling of the " "sinogram to be displayed"
        )

    def loadSinogram(self):
        """simulate the click on `load` of the QPushButton"""
        self._loadButton.clicked.emit()

    def getOptionsDockWidget(self):
        return self._dockOpt

    def _updatePlot(self, sinogram):
        super()._updatePlot(sinogram)
        if (
            self._min is not None
            and self._max is not None
            and self._estimated_cor is not None
        ):
            self.setCorRange(
                cor_min=self._min,
                cor_max=self._max,
                estimated_cor=self._estimated_cor,
                other_cors=self._other_cors,
            )

    def setCorRange(self, cor_min, cor_max, estimated_cor, other_cors=tuple()):
        """
        For now we can only modify the estimated cor

        :param float min:
        :param float max:
        :param float estimated_cor:
        :param int display_n_recons: the grid created by all cor observe
        :return:
        """
        self._min = cor_min
        self._max = cor_max
        self._estimated_cor = estimated_cor
        self._other_cors = other_cors if other_cors is not None else tuple()

        plot = self._plot

        if self._corMarker is not None:
            self._corMarker.sigDragFinished.disconnect(self._middleMarkerMoved)
        if self._fromMarker is not None:
            self._fromMarker.sigDragFinished.disconnect(
                self._fromCallback  # pylint: disable=E0203
            )
        if self._toMarker is not None:
            self._toMarker.sigDragFinished.disconnect(
                self._toCallback  # pylint: disable=E0203
            )

        # clear existing markers
        markers = [
            item.getName()
            for item in plot.getItems()
            if isinstance(item, items.MarkerBase)
        ]
        for marker in markers:
            plot.removeMarker(marker)
        # recreate markers
        plot.addXMarker(
            x=cor_min, legend="from", text="from", color="#3449ebff", draggable=True
        )
        plot.addXMarker(
            x=cor_max, legend="to", text="to", color="#3449ebff", draggable=True
        )
        plot.addXMarker(
            x=estimated_cor,
            legend="estimated cor",
            text="estimated cor",
            color="#34dcebff",
            draggable=True,
        )
        for i_cor, cor in enumerate(other_cors):
            plot.addXMarker(
                x=cor, legend=f"cor_{i_cor}", draggable=False, color="#34dceb42"
            )

        # handle draggable markers
        self._fromMarker = plot._getMarker("from")
        self._fromCallback = functools.partial(self._limitMarkerMoved, self._fromMarker)
        self._fromMarker.sigDragFinished.connect(self._fromCallback)

        self._toMarker = plot._getMarker("to")
        self._toCallback = functools.partial(self._limitMarkerMoved, self._toMarker)
        self._toMarker.sigDragFinished.connect(self._toCallback)

        self._corMarker = plot._getMarker("estimated cor")
        self._corMarker.sigDragFinished.connect(self._middleMarkerMoved)

    def _middleMarkerMoved(self):
        if self._corMarker is not None:
            self.sigCorEdited.emit(self._corMarker.getXPosition())

    def _limitMarkerMoved(self, marker):
        if self._corMarker is not None and marker is not None:
            new_half_width = numpy.abs(
                marker.getXPosition() - self._corMarker.getXPosition()
            )
            self.sigWidthEdited.emit(new_half_width * 2.0)


class _ReconstructionModeGB(qt.QGroupBox):
    reconstructionModeChanged = qt.Signal(str)

    reconstructionSliceChanged = qt.Signal()

    _DEFAULT_VERTICAL_SLICE_MODE = ("middle", "other")

    def __init__(self, parent=None, title="Slice to be reconstructed"):
        self.__scan = None
        qt.QGroupBox.__init__(self, parent)
        self.setTitle(title)
        self.setLayout(qt.QVBoxLayout())
        self._verticalRB = qt.QRadioButton(ReconstructionMode.VERTICAL.value, self)
        self.layout().addWidget(self._verticalRB)
        self._tiltCorrectionRB = qt.QRadioButton(
            ReconstructionMode.TILT_CORRECTION.value, self
        )
        self.layout().addWidget(self._tiltCorrectionRB)

        # slice choice combobox
        self._defaultSlicesCB = qt.QComboBox(self)
        for mode in self._DEFAULT_VERTICAL_SLICE_MODE:
            self._defaultSlicesCB.addItem(mode)
        self.layout().addWidget(self._defaultSlicesCB)

        # slice selector
        self._sliceVerticalQSB = SliceSelector(self, insert=False, invert_y_axis=True)
        self._sliceVerticalQSB.addSlice(value=0, name="Slice", color="green")
        self._sliceVerticalQSB.setFixedSize(qt.QSize(250, 250))
        self.layout().addWidget(self._sliceVerticalQSB)
        self._sliceTiltCorrQSB = SliceSelector(self, insert=False, invert_y_axis=True)
        self._sliceTiltCorrQSB.addSlice(value=0, name="Slice 0", color="#7e8dff")
        self._sliceTiltCorrQSB.addSlice(value=0, name="Slice 1", color="cyan")
        self._sliceTiltCorrQSB.setFixedSize(qt.QSize(250, 230))
        self.layout().addWidget(self._sliceTiltCorrQSB)
        # select on radio
        # line selector button
        onRadioIcon = silxicons.getQIcon("window-new")
        self._onRadioButton = qt.QPushButton(
            onRadioIcon, "select on radio", parent=self
        )
        self._onRadioButton.setToolTip("Open a dialog to pick the rows from" "a radio")
        self.layout().addWidget(self._onRadioButton)

        # connect signal / slot
        self._verticalRB.toggled.connect(self._reconstructionModeChanged)
        self._tiltCorrectionRB.toggled.connect(self._reconstructionModeChanged)
        self._sliceVerticalQSB.sigSlicesChanged.connect(self.reconstructionSliceChanged)
        self._onRadioButton.released.connect(self._onRadioActivated)
        self._defaultSlicesCB.currentIndexChanged.connect(self._updateVerticalSelection)

        # set up
        self.blockSignals(True)
        self.setCurrentMode("vertical", "middle")
        self.blockSignals(False)
        self._tiltCorrectionRB.hide()
        # as for now only the Vertical mode is handled hide it to avoid confusing user
        self._verticalRB.hide()

    def setCurrentMode(self, mode, slices):
        valid_modes = ("vertical", "horizontal")
        if mode not in valid_modes:
            raise ValueError(
                f"mode {mode} is not recosgnized. Valid modes are {valid_modes}"
            )
        self._verticalRB.setChecked(mode == "vertical")
        if mode == "vertical":
            if slices not in self._DEFAULT_VERTICAL_SLICE_MODE:
                raise ValueError(
                    f"{slices} is not a vadid slice mode. Calid are {self._DEFAULT_VERTICAL_SLICE_MODE}"
                )
            else:
                index = self._defaultSlicesCB.findText(slices)
                self._defaultSlicesCB.setCurrentIndex(index)
                self._defaultSlicesCB.currentIndexChanged.emit(index)

    def _updateVerticalSelection(self):
        current_mode = self._defaultSlicesCB.currentText()
        mode_is_middle = current_mode == "middle"
        self._sliceVerticalQSB.setVisible(not mode_is_middle)
        self._onRadioButton.setVisible(not mode_is_middle)
        self.reconstructionSliceChanged.emit()

    def _onRadioActivated(self):
        lineSelection = QSliceSelectorDialog(parent=self, n_required_slice=1)
        scan = self.__scan() if self.__scan else None
        if scan is None:
            _logger.info("No scan defined, request user for defining the image" "url")
            dialog = ImageFileDialog()
            if dialog.exec_():
                data = dialog.selectedImage()
            else:
                return
        else:
            assert isinstance(scan, TomwerScanBase)
            try:
                projections = scan.projections
                index_radio = list(projections.keys())[len(projections) // 2]
                data = get_data(projections[index_radio])
            except Exception as e:
                _logger.error(f"Fail to load radio data: {str(e)}")
                return

        lineSelection.setData(data)
        lineSelection.setSelection(list(self.getReconstructionSlices().values()))
        if lineSelection.exec_() == qt.QDialog.Accepted:
            selection = lineSelection.getSelection()
            if selection is not None and len(selection) > 0:
                self.setReconstructionSlices({"Slice": selection[0]})

    def setScan(self, scan):
        self.__scan = weakref.ref(scan)

    def getReconstructionMode(self) -> ReconstructionMode:
        if self._tiltCorrectionRB.isChecked():
            return ReconstructionMode.TILT_CORRECTION
        elif self._verticalRB.isChecked():
            return ReconstructionMode.VERTICAL
        else:
            raise ValueError("No reconstruction mode selected")

    def setReconstructionMode(self, mode: Union[str, ReconstructionMode]) -> None:
        mode = ReconstructionMode.from_value(mode)
        if mode is ReconstructionMode.TILT_CORRECTION:
            self._tiltCorrectionRB.setChecked(True)
        elif mode is ReconstructionMode.VERTICAL:
            self._verticalRB.setChecked(True)
        else:
            raise ValueError("No reconstruction mode selected")

    def getReconstructionSlices(self):
        mode = self.getReconstructionMode()
        if mode is ReconstructionMode.TILT_CORRECTION:
            return self._sliceTiltCorrQSB.getSlicesValue()
        elif mode is ReconstructionMode.VERTICAL:
            if self._defaultSlicesCB.currentText() == "middle":
                return "middle"
            else:
                return self._sliceVerticalQSB.getSlicesValue()
        else:
            raise ValueError(f"mode {mode} is not handled")

    def setReconstructionSlices(self, slice_: Union[dict, int, str]):
        slice_name = None
        if slice_ == "middle" or isinstance(slice_, int) or len(slice_) == 1:
            self.setReconstructionMode(ReconstructionMode.VERTICAL)
            target = self._sliceVerticalQSB
            if slice_ == "middle":
                index = self._defaultSlicesCB.findText("middle")
            else:
                index = self._defaultSlicesCB.findText("other")
                slice_name = "Slice"
            self._defaultSlicesCB.setCurrentIndex(index)
        elif len(slice_) == 2:
            self.setReconstructionMode(ReconstructionMode.TILT_CORRECTION)
            target = self._sliceTiltCorrQSB
        else:
            raise ValueError("slice is expected to contain one or two element(s)")
        if slice_ not in ("middle",):
            if isinstance(slice_, int):
                target.setSliceValue(name=slice_name, value=slice_)
            else:
                for slice_name, slice_value in slice_.items():
                    target.setSliceValue(name=slice_name, value=slice_value)

    def _reconstructionModeChanged(self, *args, **kwargs):
        reconstruction_mode = self.getReconstructionMode()
        self.reconstructionModeChanged.emit(reconstruction_mode.value)
        self._sliceTiltCorrQSB.setVisible(
            reconstruction_mode is ReconstructionMode.TILT_CORRECTION
        )
        self._sliceVerticalQSB.setVisible(
            reconstruction_mode is ReconstructionMode.VERTICAL
        )

    def setSlicesRange(self, min_index, max_index):
        self._sliceTiltCorrQSB.setSlicesRange(min_index, max_index)
        self._sliceVerticalQSB.setSlicesRange(min_index, max_index)

    def getSlicesRange(self):
        return self._sliceVerticalQSB.getSlicesRange()


class _DetectionAccuracyGB(qt.QGroupBox):
    detectionAccuracyChanged = qt.Signal()
    """signal emitted when the value change"""

    _MIN_RANGE = 0.1
    _MAX_RANGE = 4000
    _DEFAULT_VALUE = 10

    def __init__(self, parent=None, title="Detection accuracy (width)"):
        qt.QGroupBox.__init__(self, parent)
        self.setTitle(title)
        self.setLayout(qt.QGridLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self.setContentsMargins(0, 0, 0, 0)
        # slider
        self._accuracySlider = LogSlider(parent=self)
        self._accuracySlider.setSuffix(" px")
        self.layout().addWidget(self._accuracySlider, 0, 0, 1, 5)
        # different accuracy buttons
        accuracy_opt_font = self.font()
        accuracy_opt_font.setPixelSize(10)
        self._coarsePB = qt.QPushButton(DetectionAccuracy.COARSE.value, self)
        self._coarsePB.setFont(accuracy_opt_font)
        self._coarsePB.setFlat(True)
        self.layout().addWidget(self._coarsePB, 1, 2, 1, 1)
        self._finePB = qt.QPushButton(DetectionAccuracy.FINE.value, self)
        self._finePB.setFont(accuracy_opt_font)
        self._finePB.setFlat(True)
        self.layout().addWidget(self._finePB, 1, 1, 1, 1)
        self._accuratePB = qt.QPushButton(DetectionAccuracy.ACCURATE.value, self)
        self._accuratePB.setFont(accuracy_opt_font)
        self._accuratePB.setFlat(True)
        self.layout().addWidget(self._accuratePB, 1, 0, 1, 1)

        # connect signal / slot
        self._coarsePB.released.connect(self._setCoarse)
        self._finePB.released.connect(self._setFine)
        self._accuratePB.released.connect(self._setAccurate)
        self._accuracySlider.valueChanged.connect(self.detectionAccuracyChanged)

        # set up
        self._accuracySlider.setRange(self._MIN_RANGE, self._MAX_RANGE)
        self._accuracySlider.setValue(self._DEFAULT_VALUE)

    def _detectionAccuracyChanged(self, *args, **kwargs):
        self.detectionAccuracyChanged.emit()

    def getResearchWidth(self) -> float:
        """Return research width (in pixel)"""
        return self._accuracySlider.value()

    def setResearchWidth(self, width: float):
        """set research width

        :param float width: center of rotation research width
        """
        return self._accuracySlider.setValue(width)

    def _setCoarse(self):
        self._accuracySlider.setValue(ACCURACIES[DetectionAccuracy.COARSE])

    def _setFine(self):
        self._accuracySlider.setValue(ACCURACIES[DetectionAccuracy.FINE])

    def _setAccurate(self):
        self._accuracySlider.setValue(ACCURACIES[DetectionAccuracy.ACCURATE])


class _EstimatedCorWidget(qt.QGroupBox):
    _MIDDLE_COR_TXT = "middle"
    _OTHER_COR_TXT = "other"

    sigAutoCorRequested = qt.Signal()
    """signal emitted when the calculation of automatic cor is asked"""

    sigAutoCorChanged = qt.Signal()
    """signal emitted when the estimated cor value change"""

    def __init__(self, parent=None, title=None):
        assert title is not None
        self._frameWidth = None
        qt.QGroupBox.__init__(self, parent)
        self.setTitle(title)
        self.setLayout(qt.QGridLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self._middlePositionRB = qt.QCheckBox("middle", self)
        self.layout().addWidget(self._middlePositionRB, 0, 0, 1, 2)
        # manual definition of the cor
        self._manualCORAbs = qt.QDoubleSpinBox(self)
        self._manualCORAbs.setDecimals(4)
        self._manualCORAbs.setMaximum(9999999)
        self._manualCORAbs.setSuffix(" px")
        self._absLabel = qt.QLabel("absolute: ")
        self.layout().addWidget(self._absLabel, 1, 0, 1, 1)
        self.layout().addWidget(self._manualCORAbs, 1, 1, 1, 1)
        self._manualCORRel = qt.QDoubleSpinBox(self)
        self._manualCORRel.setDecimals(4)
        self._manualCORRel.setSuffix(" px")
        self._manualCORRel.setMinimum(-9999999)
        self._manualCORRel.setMaximum(9999999)
        self._relLabel = qt.QLabel("relative: ")
        self.layout().addWidget(self._relLabel, 2, 0, 1, 1)
        self.layout().addWidget(self._manualCORRel, 2, 1, 1, 1)
        # auto cor button
        self._autoCorPB = qt.QPushButton("auto cor", self)
        auto_cor_icon = icons.getQIcon("a")
        self._autoCorPB.setIcon(auto_cor_icon)
        self.layout().addWidget(self._autoCorPB, 3, 1, 1, 1)

        # connect signal / slot
        self._middlePositionRB.toggled.connect(self._modeChanged)
        self._manualCORRel.editingFinished.connect(self._relativePositionChanged)
        self._manualCORAbs.editingFinished.connect(self._absolutePositionChanged)
        self._autoCorPB.released.connect(self.sigAutoCorRequested)
        self._middlePositionRB.toggled.connect(self.sigAutoCorChanged)
        self._manualCORRel.editingFinished.connect(self.sigAutoCorChanged)
        self._manualCORAbs.editingFinished.connect(self.sigAutoCorChanged)
        # set up
        self._middlePositionRB.setChecked(True)
        self._modeChanged()

    def getCorValue(self, mode="relative"):
        assert mode in ("rel", "relative", "abs", "absolute")
        if self._middlePositionRB.isChecked():
            if mode in ("rel", "relative"):
                return 0.0
            elif mode in ("abs", "absolute"):
                if self._frameWidth is not None:
                    return self._frameWidth / 2.0
                else:
                    return self._MIDDLE_COR_TXT
            else:
                raise ValueError(f"Mode {mode} is not recognized")
        else:
            if mode in ("rel", "relative"):
                return self._manualCORRel.value()
            elif mode in ("abs", "absolute"):
                return self._manualCORAbs.value()
            else:
                raise ValueError(f"Mode {mode} is not recognized")

    def _modeChanged(self, *args, **kwargs):
        manual_opt_visible = not self._middlePositionRB.isChecked()
        self._absLabel.setVisible(manual_opt_visible)
        self._relLabel.setVisible(manual_opt_visible)
        self._manualCORRel.setVisible(manual_opt_visible)
        self._manualCORAbs.setVisible(manual_opt_visible)
        self._autoCorPB.setVisible(manual_opt_visible)

    def _relativePositionChanged(self, *args, **kwargs):
        if self._frameWidth is not None:
            old = self.blockSignals(True)
            old_mcor = self._manualCORAbs.blockSignals(True)
            self._manualCORAbs.setValue(
                self._manualCORRel.value() + self._frameWidth / 2.0
            )
            self._manualCORAbs.blockSignals(old_mcor)
            self.blockSignals(old)

    def _absolutePositionChanged(self, *args, **kwargs):
        if self._frameWidth is not None:
            old = self.blockSignals(True)
            old_mcor = self._manualCORRel.blockSignals(True)
            self._manualCORRel.setValue(
                self._manualCORAbs.value() - self._frameWidth / 2.0
            )
            self._manualCORRel.blockSignals(old_mcor)
            self.blockSignals(old)

    def setFrameWidth(self, width):
        self._frameWidth = width
        # if necessary auto reset of the absolute value
        if self._manualCORAbs.value() == self._manualCORRel.value():
            self._relativePositionChanged()

    def getFrameWidth(self):
        return self._frameWidth

    def setCorValue(self, value):
        """

        :param Union[str,float] value: if a float expected to be given as the
                                       relative value
        :return:
        """
        if value is None:
            pass
        elif value == self._MIDDLE_COR_TXT:
            self._middlePositionRB.setChecked(True)
        else:
            self._middlePositionRB.setChecked(False)
            self._manualCORRel.setValue(float(value))
            self._manualCORRel.editingFinished.emit()

    def setAbsoluteCorValue(self, value):
        if value is None:
            pass
        elif value == self._MIDDLE_COR_TXT:
            self._middlePositionRB.setChecked(True)
        else:
            self._middlePositionRB.setChecked(False)
            self._manualCORAbs.setValue(float(value))
            self._manualCORAbs.editingFinished.emit()


class SAAxisOptions(qt.QWidget):
    """
    Options of the semi-automatic center of rotation detection
    """

    sigReconstructionSliceChanged = qt.Signal()
    """signal emitted when the slice to reconstruct change"""

    sigReconstructionRangeChanged = qt.Signal()
    """signal emitted when the reconstruction parameters change"""

    sigConfigurationChanged = qt.Signal()
    """signal emitted when the configuration change"""

    _FONT_IMPORTANT = qt.QFont("Arial", 12)
    _FONT_MEDIUM = qt.QFont("Arial", 10)

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        # reconstruction mode
        self._reconstructionMode = _ReconstructionModeGB(parent=self)
        self._reconstructionMode.setFont(self._FONT_MEDIUM)
        self.layout().addWidget(self._reconstructionMode)

        # estimated cor
        self._estimatedCorWidget = _EstimatedCorWidget(
            self,
            title="Estimated cor position (x axis)",
        )
        self._estimatedCorWidget.setFont(self._FONT_IMPORTANT)
        self.layout().addWidget(self._estimatedCorWidget)

        # detection accuracy
        self._detectionAccuracy = _DetectionAccuracyGB(parent=self)
        self._detectionAccuracy.setFont(self._FONT_MEDIUM)
        self.layout().addWidget(self._detectionAccuracy)

        # number of reconstructions eq volume size
        self._nReconsWidget = qt.QWidget(parent)
        self._nReconsWidget.setLayout(qt.QFormLayout())
        self._nReconsWidget.layout().setContentsMargins(0, 0, 0, 0)
        self._nReconsSB = qt.QSpinBox(self._nReconsWidget)
        self._nReconsSB.setRange(0, 2000)
        self._nReconsSB.setValue(30)
        self._nReconsWidget.setFont(self._FONT_IMPORTANT)
        self._nReconsWidget.layout().addRow("Number of reconstruction", self._nReconsSB)
        self.layout().addWidget(self._nReconsWidget)

        # spacer
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        # expose API
        self.sigAutoCorRequested = self._estimatedCorWidget.sigAutoCorRequested

        # connect signal / slot
        self._reconstructionMode.reconstructionSliceChanged.connect(
            self.sigReconstructionSliceChanged
        )
        self._nReconsSB.valueChanged.connect(self._reconstructionRangeChanged)
        self._estimatedCorWidget.sigAutoCorChanged.connect(
            self._reconstructionRangeChanged
        )
        self._detectionAccuracy.detectionAccuracyChanged.connect(
            self._reconstructionRangeChanged
        )

    def setDetectionAccuracy(self, width):
        if width != self._detectionAccuracy.getResearchWidth():
            self._detectionAccuracy.setResearchWidth(width)
            self.sigConfigurationChanged.emit()

    def setEstimatedCorPosition(self, value):
        if value != self._estimatedCorWidget.getCorValue("relative"):
            self._estimatedCorWidget.setCorValue(value)
            self.sigConfigurationChanged.emit()

    def setEstimatedAbsoluteCorPosition(self, value):
        if value != self._estimatedCorWidget.getCorValue("absolute"):
            self._estimatedCorWidget.setAbsoluteCorValue(value)
            self.sigConfigurationChanged.emit()

    def getEstimatedCorPosition(self, *args, **kwargs):
        return self._estimatedCorWidget.getCorValue(*args, **kwargs)

    def setScan(self, scan: TomwerScanBase):
        if scan is not None:
            self.setSlicesRange(0, scan.dim_2)
            self._estimatedCorWidget.setFrameWidth(scan.dim_1)
            self._reconstructionMode.setScan(scan)

    def getFrameWidth(self):
        return self._estimatedCorWidget.getFrameWidth()

    def setFrameWidth(self, width):
        if width != self._estimatedCorWidget.getFrameWidth():
            self._estimatedCorWidget.setFrameWidth(width)
            self.sigConfigurationChanged.emit()

    def getSlicesRange(self) -> tuple:
        self._reconstructionMode.getSlicesRange()

    def setSlicesRange(self, min_index, max_index):
        self._reconstructionMode.setSlicesRange(min_index, max_index)
        self.sigConfigurationChanged.emit()

    def _reconstructionRangeChanged(self):
        self.sigReconstructionRangeChanged.emit()
        self.sigConfigurationChanged.emit()

    def getReconstructionSlices(self) -> tuple:
        return self._reconstructionMode.getReconstructionSlices()

    def setReconstructionSlices(self, slices) -> tuple:
        return self._reconstructionMode.setReconstructionSlices(slices)

    def getReconstructionMode(self):
        return self._reconstructionMode.getReconstructionMode()

    def setReconstructionMode(self, mode):
        self._reconstructionMode.setReconstructionMode(mode)
        self.sigConfigurationChanged.emit()

    def getNReconstruction(self):
        return self._nReconsSB.value()

    def setNReconstruction(self, n):
        self._nReconsSB.setValue(int(n))
        self.sigConfigurationChanged.emit()

    def getResearchWidth(self):
        return self._detectionAccuracy.getResearchWidth()

    def setResearchWidth(self, width):
        self._detectionAccuracy.setResearchWidth(width)
        self.sigConfigurationChanged.emit()


class SliceAndCorWidget(qt.QMainWindow):
    sigConfigurationChanged = qt.Signal()
    """signal emitted when the configuration changed"""

    def __init__(self, parent=None):
        qt.QMainWindow.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)
        # sinogram viewer
        self._sinogramViewer = SinogramViewer(self)
        self._sinogramViewer.setWindowFlags(qt.Qt.Widget)
        self.setCentralWidget(self._sinogramViewer)
        # slice and accuracy options
        self._saaxisOpts = SAAxisOptions(self)
        self._dockWidgetOpts = qt.QDockWidget(self)
        self._dockWidgetOpts.setFixedWidth(300)
        self._dockWidgetOpts.setWidget(self._saaxisOpts)
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self._dockWidgetOpts)
        self._dockWidgetOpts.setFeatures(qt.QDockWidget.DockWidgetMovable)
        # sinogram options
        self.sinoOprsDW = self._sinogramViewer.getOptionsDockWidget()
        self.addDockWidget(qt.Qt.RightDockWidgetArea, self.sinoOprsDW)
        self.sinoOprsDW.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self.sinoOprsDW.setFixedWidth(300)

        # expose API
        self.getEstimatedCorPosition = self._saaxisOpts.getEstimatedCorPosition
        self.setEstimatedCorPosition = self._saaxisOpts.setEstimatedCorPosition
        self.setEstimatedAbsoluteCorPosition = (
            self._saaxisOpts.setEstimatedAbsoluteCorPosition
        )
        self.setDetectionAccuracy = self._saaxisOpts.setDetectionAccuracy
        self.getNReconstruction = self._saaxisOpts.getNReconstruction
        self.setNReconstruction = self._saaxisOpts.setNReconstruction
        self.getResearchWidth = self._saaxisOpts.getResearchWidth
        self.setResearchWidth = self._saaxisOpts.setResearchWidth
        self.getReconstructionSlices = self._saaxisOpts.getReconstructionSlices
        self.setReconstructionSlices = self._saaxisOpts.setReconstructionSlices
        self.getReconstructionMode = self._saaxisOpts.getReconstructionMode
        self.setReconstructionMode = self._saaxisOpts.setReconstructionMode
        self.getFrameWidth = self._saaxisOpts.getFrameWidth
        self.setFrameWidth = self._saaxisOpts.setFrameWidth
        self.getSlicesRange = self._saaxisOpts.getSlicesRange
        self.setSlicesRange = self._saaxisOpts.setSlicesRange
        self.sigStartSinogramLoad = self._sinogramViewer.sigSinoLoadStarted
        self.sigEndSinogramLoad = self._sinogramViewer.sigSinoLoadEnded
        self.sigConfigurationChanged = self._saaxisOpts.sigConfigurationChanged

        # expose signals
        self.sigReconstructionSliceChanged = (
            self._saaxisOpts.sigReconstructionSliceChanged
        )
        self.sigAutoCorRequested = self._saaxisOpts.sigAutoCorRequested
        self.sigReconstructionRangeChanged = (
            self._saaxisOpts.sigReconstructionRangeChanged
        )

        # connect signal / slot
        self._sinogramViewer.sigCorEdited.connect(self.setEstimatedAbsoluteCorPosition)
        self._sinogramViewer.sigWidthEdited.connect(self.setDetectionAccuracy)

    def loadSinogram(self):
        """Enforce to load a sinogram with the current settings"""
        self._sinogramViewer.loadSinogram()

    def setScan(self, scan):
        self._sinogramViewer.setScan(scan, update=False)
        self._saaxisOpts.setScan(scan)

    def close(self):
        self._sinogramViewer.close()
        self._sinogramViewer = None
        super().close()



import os
from tomoscan.serie import Serie

from tomwer.core.process.control.nxtomoconcatenate import (
    ConcatenateNXtomoTask,
    format_output_location,
)
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.utils.scanutils import MockNXtomo


def test_concatenate_nx_tomo_task(tmp_path):
    """
    test execution of ConcatenateNXtomoTask with two NXtomoScan
    """
    scan_1 = MockNXtomo(
        scan_path=os.path.join(tmp_path, "scan1"),
        n_proj=10,
        n_ini_proj=10,
        scan_range=180,
        dim=20,
        energy=12.3,
    ).scan
    scan_2 = MockNXtomo(
        scan_path=os.path.join(tmp_path, "scan2"),
        n_proj=10,
        n_ini_proj=10,
        scan_range=180,
        dim=20,
        energy=12.3,
    ).scan

    output_scan_file = os.path.join(tmp_path, "concatenation.nx")
    assert not os.path.exists(output_scan_file)
    task = ConcatenateNXtomoTask(
        inputs={
            "serie": Serie(iterable=[scan_1, scan_2]),
            "output_file": output_scan_file,
            "output_entry": "my_entry",
            "overwrite": False,
            "serialize_output_data": False,
        }
    )
    task.run()
    assert isinstance(task.outputs.data, NXtomoScan)
    assert os.path.exists(output_scan_file)
    # note: correct processing of the concatenation is done at nxtomomill


def test_format_output_location():
    """test behavior of format_output_location"""
    scan1 = NXtomoScan(
        "/path/to/scan_1.nx",
        entry="entry0000",
    )
    scan2 = NXtomoScan(
        "/path/to/scan_2.nx",
        entry="entry0001",
    )
    assert (
        format_output_location(
            "{common_path}/concatenate.nx", Serie(iterable=[scan1, scan2])
        )
        == "/path/to/concatenate.nx"
    )

    assert format_output_location(
        "concatenate.nx", Serie("my_serie", [scan1, scan2])
    ) == os.path.abspath("concatenate.nx")

    assert (
        format_output_location(
            "{common_path}/concatenate.nx",
            Serie(
                iterable=[
                    scan1,
                ]
            ),
        )
        == "/path/to/concatenate.nx"
    )



"""
Some utils GUI associated to illustrations
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "29/01/2020"


import logging
import os
import subprocess
import threading
import platform

_logger = logging.getLogger(__name__)


__has_image_j = None
__image_j_from = None

_IMAGE_J_BIN_PATH = "/sware/pub/ImageJ/ImageJ"


def has_imagej():
    """Return if imagej command is accessible from the computer"""
    global __has_image_j
    global __image_j_from
    if __has_image_j is None:
        if platform.machine() in ("x86_64", "AMD64") and os.path.exists(
            _IMAGE_J_BIN_PATH
        ):
            __image_j_from = "binary"
            __has_image_j = True
        else:
            try:
                # use help because there is no information regarding version
                subprocess.call(["imagej", "-h"], stdout=subprocess.PIPE)
            except Exception:
                __has_image_j = False
            else:
                __image_j_from = "command"
                __has_image_j = True
    return __has_image_j


def open_url_with_image_j(url) -> threading.Thread:
    """open the url in an imagej subprocess within a thread.
    It is up to the caller to handle thread life

    :param DataUrl url: url we want to open in imagej
    """
    global __image_j_from
    if not has_imagej():
        raise OSError("ImageJ is not installed")
    thread = ImageJthread(url=url, image_j_from=__image_j_from)
    thread.start()
    return thread


class ImageJthread(threading.Thread):
    def __init__(self, url, image_j_from, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._image_j_from = image_j_from
        self._url = url
        self._process = None

    def run(self):
        if self._image_j_from == "command":
            # for now we only manage the simple case of an edf file
            try:
                self._process = subprocess.Popen(
                    ["imagej", "-o", self._url.file_path()],
                )
            except Exception as e:
                _logger.warning(f"Fail to open {self._url}. Reason is {e}")
            else:
                self._process.communicate()

        elif self._image_j_from == "binary":
            try:
                self._process = subprocess.Popen(
                    [_IMAGE_J_BIN_PATH, self._url.file_path()],
                )
            except Exception as e:
                _logger.warning(f"Fail to open {self._url}. Reason is {e}")
            else:
                self._process.communicate()

    def quit(self):
        if self._process is not None:
            self._process.kill()
            self._process.wait()



try:
    from nabu.pipeline.fullfield.reconstruction import (  # noqa F401
        FullFieldReconstructor,
    )
except (ImportError, OSError):
    try:
        from nabu.pipeline.fullfield.local_reconstruction import (  # noqa F401
            ChunkedReconstructor,
        )
    except (ImportError, OSError):
        # import of cufft library can bring an OSError if cuda not install
        has_nabu = False
    else:
        has_nabu = True
else:
    has_nabu = True
import logging
import os
import sys
from copy import deepcopy
import subprocess
from typing import Iterable, Optional, Union
import numpy

from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from processview.core.manager.manager import ProcessManager
from sluurp.job import SBatchScriptJob
from sluurp.executor import submit as submit_to_slurm_cluster

from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.process.reconstruction.nabu.nabuslices import (
    SingleSliceRunner,
    _NabuBaseReconstructor,
    generate_nabu_configfile,
)
from tomwer.core.process.reconstruction.nabu.utils import (
    slice_index_to_int,
    get_nabu_multicor_file_prefix,
)
from tomwer.core.process.reconstruction.nabu.target import Target
from tomwer.core.progress import Progress
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.slurm import get_slurm_script_name, is_slurm_available
from tomwer.utils import docstring
from tomwer.core.process.reconstruction.nabu.utils import nabu_std_err_has_error
from tomwer.core.process.reconstruction.nabu.plane import NabuPlane

from ..nabu import settings as nabu_settings
from . import utils, settings
from .nabucommon import ResultsLocalRun, ResultSlurmRun, ResultsWithStd, ResultsRun

_logger = logging.getLogger(__name__)


def run_nabu_one_slice_several_config(
    scan: TomwerScanBase,
    nabu_configs: Union[list, tuple],
    cluster_config: Optional[dict],
    dry_run: bool,
    slice_index: Union[int, str],
    file_format: str,
    axis: NabuPlane,
    advancement: Optional[Progress] = None,
    process_id: Optional[int] = None,
    instanciate_classes_only: bool = False,
    output_file_prefix_pattern=None,
) -> tuple:
    """
    Run several reconstruction of a specific slice.

    :param scan: dataset
    :param Iterable nabu_configs: set of nabu configurations to be run
    :param bool dry_run:
    :param int slice_index: slice index to reconstruct or "middle"
    :param Progress advancement: optional class to display advancement
    :param Optional[int] process_id: id of the process requesting this computation
    :param Optional[dict] cluster_config: cluster configuration if
    :return: success, recons_urls (list of output urls), tuple of outs, tuples of errs, dict future_scans (key is cor, value is future_scan)
             if `instanciate_classes_only` set to True then return a list of :class:`_Reconstructor`
    :rtype: tuple
    """
    if cluster_config in (None, {}):
        target = Target.LOCAL
    elif isinstance(cluster_config, dict):
        if not is_slurm_available():
            raise RuntimeError("Slurm computation requested but unvailable")
        target = Target.SLURM
    else:
        raise TypeError(
            f"cluster_config should be None or a dict not {type(cluster_config)}"
        )

    if process_id is not None:
        try:
            process_name = ProcessManager().get_process(process_id=process_id).name
        except KeyError:
            process_name = "unknow"
    else:
        process_name = ""

    reconstructor = _Reconstructor(
        scan=scan,
        nabu_configs=nabu_configs,
        advancement=advancement,
        slice_index=slice_index,
        target=target,
        dry_run=dry_run,
        file_format=file_format,
        cluster_config=cluster_config,
        process_name=process_name,
        output_file_prefix_pattern=output_file_prefix_pattern,
        axis=axis,
    )
    if instanciate_classes_only:
        return (reconstructor,)

    try:
        results = reconstructor.run()
    except TimeoutError as e:
        _logger.error(e)
        return None
    else:
        assert isinstance(
            results, dict
        ), "results should be a dictionary with var_value as key and urls as value"
        success = True
        recons_urls = {}
        std_outs = []
        std_errs = []
        future_tomo_objs = {}
        for var_value, res in results.items():
            success = success and res.success
            if isinstance(res, ResultsWithStd):
                std_outs.append(res.std_out)
                std_errs.append(res.std_err)
            if isinstance(res, ResultsLocalRun):
                recons_urls[var_value] = res.results_identifiers
            if isinstance(res, ResultSlurmRun):
                future_tomo_obj = FutureTomwerObject(
                    tomo_obj=scan,
                    process_requester_id=process_id,
                    futures=res.future_slurm_jobs,
                )
                future_tomo_objs[var_value] = future_tomo_obj
        return success, recons_urls, std_outs, std_errs, future_tomo_objs


def run_nabu_multicor(
    scan: TomwerScanBase,
    nabu_config: dict,
    cors: tuple,
    cluster_config: Optional[dict],
    dry_run: bool,
    slice_index: Union[int, str],
    file_format: str,
    process_id: Optional[int] = None,
    instanciate_classes_only: bool = False,
    output_file_prefix_pattern=None,
):
    if cluster_config in (None, {}):
        target = Target.LOCAL
    elif isinstance(cluster_config, dict):
        if not is_slurm_available():
            raise RuntimeError("Slurm computation requested but unvailable")
        target = Target.SLURM
    else:
        raise TypeError(
            f"cluster_config should be None or a dict not {type(cluster_config)}"
        )

    if process_id is not None:
        try:
            process_name = ProcessManager().get_process(process_id=process_id).name
        except KeyError:
            process_name = "unknow"
    else:
        process_name = ""

    # TODO: FIXME small hack to make sure the configuration will be accepted when valie
    # for now even if the cor values are given from a dedicated parameter nabu is still
    # checking the value provided in the config file. If this value is invalid for
    # half acquisition it will be refused
    nabu_config["reconstruction"]["rotation_axis_position"] = numpy.mean(cors)

    axis = nabu_config["reconstruction"].get("slice_plane", "XY")

    reconstructor = _ReconstructorMultiCor(
        scan=scan,
        nabu_config=nabu_config,
        cors=cors,
        slice_index=slice_index,
        target=target,
        dry_run=dry_run,
        file_format=file_format,
        cluster_config=cluster_config,
        process_name=process_name,
        axis=axis,
        output_file_prefix_pattern=output_file_prefix_pattern,
    )
    if instanciate_classes_only:
        return reconstructor

    try:
        result = reconstructor.run()
    except TimeoutError as e:
        _logger.error(e)
        return None
    else:
        recons_urls = {}
        std_outs = []
        std_errs = []
        future_tomo_obj = None

        success = result.success
        if isinstance(result, ResultsWithStd):
            std_outs.append(result.std_out)
            std_errs.append(result.std_err)
        if isinstance(result, ResultsLocalRun):
            recons_urls = result.results_identifiers
        if isinstance(result, ResultSlurmRun):
            future_tomo_obj = FutureTomwerObject(
                tomo_obj=scan,
                process_requester_id=process_id,
                futures=result.future_slurm_jobs,
            )
        return success, recons_urls, (future_tomo_obj,), std_outs, std_errs


class _Reconstructor(_NabuBaseReconstructor):
    def __init__(
        self,
        scan: TomwerScanBase,
        nabu_configs: Iterable,
        advancement: Optional[Progress],
        slice_index: Union[int, str],
        axis: Union[str, NabuPlane],
        target: Target,
        dry_run: bool,
        file_format: str,
        cluster_config: Optional[dict],
        process_name: str,
        output_file_prefix_pattern=None,
    ) -> None:
        """
        :param scan: scan to reconstruct
        :param nabu_configs: all the configuration to run
        :param advancement: Progress object to notify advancement
        :param slice_index: index of the slice to reconstruct.
        :param axis: axis over which we want to do the reconstruction
        :param target: is the reconstruction is to made locally or remotly
        :param file_format: reconstructed volume file format
        :param cluster_config: cluster configuration in the case of a remote execution
        :param str extra_output_file_pattern: possible extra file name pattern like for cor we want to add 'cor_' as prefix and cor value as suffix.
                                              To make the file name unique. For delta/beta it is already forseen to be unique. For now keywords are:
                                              * file_name: default file name according to db values and dataset name
                                              * value: value of the nabu_configs keys
        """
        super().__init__(
            scan=scan,
            dry_run=dry_run,
            target=target,
            cluster_config=cluster_config,
            process_name=process_name,
            axis=axis,
        )
        if not isinstance(slice_index, (int, str)):
            raise TypeError(
                f"slice_index should be an int or a string not {type(slice_index)}"
            )
        self.advancement = advancement
        self.slice_index = slice_index
        self.nabu_configs = nabu_configs
        self.file_format = file_format
        self._output_file_prefix_pattern = output_file_prefix_pattern

    @docstring(_NabuBaseReconstructor)
    def run(self) -> Iterable:
        self.slice_index = slice_index_to_int(
            slice_index=self.slice_index, scan=self.scan
        )

        results = {}
        if self.advancement:
            self.advancement.setMaxAdvancement(len(self.nabu_configs))
        for var_value, config in self.nabu_configs.items():
            if self._cancelled:
                break
            config, conf_file = self.preprocess_config(deepcopy(config), var_value)

            # add some tomwer metadata and save the configuration
            # note: for now the section is ignored by nabu but shouldn't stay that way
            with utils.TomwerInfo(config) as config_to_dump:
                generate_nabu_configfile(
                    conf_file,
                    nabu_fullfield_default_config,
                    config=config_to_dump,
                    options_level="advanced",
                )

            results[var_value] = self._process_config(
                config_to_dump=config_to_dump,
                config_file=conf_file,
                file_format=self.file_format,
                info="nabu slice reconstruction",
                process_name=self.process_name,
            )
            # specific treatment for cor: rename output files
            if self.advancement:
                self.advancement.increaseAdvancement(1)
        return results

    def _format_file_prefix(self, file_prefix, value):
        if self._output_file_prefix_pattern is None:
            return file_prefix

        keywords = {
            "file_name": file_prefix,
            "value": value,
        }

        # filter necessary keywords
        def get_necessary_keywords():
            import string

            formatter = string.Formatter()
            return [
                field
                for _, field, _, _ in formatter.parse(self._output_file_prefix_pattern)
                if field
            ]

        requested_keywords = get_necessary_keywords()

        def keyword_needed(pair):
            keyword, _ = pair
            return keyword in requested_keywords

        keywords = dict(filter(keyword_needed, keywords.items()))
        return self._output_file_prefix_pattern.format(**keywords)

    def treateOutputConfig(self, _config, value):
        """
        - add or overwrite some parameters of the dictionary
        - create the output directory if does not exist
        """
        pag = False
        ctf = False
        db = None
        if "phase" in _config:
            phase_method = _config["phase"].get("method", "").lower()
            if phase_method in ("pag", "paganin"):
                pag = True
            elif phase_method in ("ctf",):
                ctf = True

            if "delta_beta" in _config["phase"]:
                db = round(float(_config["phase"]["delta_beta"]))
        if "output" in _config:
            file_prefix = SingleSliceRunner.get_file_basename_reconstruction(
                scan=self.scan,
                slice_index=self.slice_index,
                pag=pag,
                db=db,
                ctf=ctf,
                axis=self.axis,
            )
            file_prefix = self._format_file_prefix(file_prefix=file_prefix, value=value)
            _config["output"]["file_prefix"] = file_prefix
            assert _config["output"]["location"] not in ("", None)
            if not os.path.isdir(_config["output"]["location"]):
                os.makedirs(_config["output"]["location"])

        if "reconstruction" not in _config:
            _config["reconstruction"] = {}
        if self.axis is NabuPlane.YZ:
            _config["reconstruction"]["start_x"] = self.slice_index
            _config["reconstruction"]["end_x"] = self.slice_index
        elif self.axis is NabuPlane.XZ:
            _config["reconstruction"]["start_y"] = self.slice_index
            _config["reconstruction"]["end_y"] = self.slice_index
        elif self.axis is NabuPlane.XY:
            _config["reconstruction"]["start_z"] = self.slice_index
            _config["reconstruction"]["end_z"] = self.slice_index
        else:
            raise ValueError(
                f"self.axis has an invalid value: {self.axis} when expected to be in {NabuPlane.values()}"
            )
        return _config, file_prefix

    def preprocess_config(self, config, value) -> tuple:
        dataset_params = self.scan.get_nabu_dataset_info()
        if "dataset" in config:
            dataset_params.update(config["dataset"])
        config["dataset"] = dataset_params

        config["resources"] = utils.get_nabu_resources_desc(
            scan=self.scan, workers=1, method="local"
        )
        # force overwrite results
        if "output" not in config:
            config["output"] = {}
        config["output"].update({"overwrite_results": 1})

        config, file_prefix = self.treateOutputConfig(config, value=value)
        # the policy is to save nabu .cfg file at the same location as the
        # force overwrite results

        cfg_folder = os.path.join(
            config["output"]["location"],
            nabu_settings.NABU_CFG_FILE_FOLDER,
        )
        os.makedirs(cfg_folder, exist_ok=True)

        conf_file = os.path.join(
            cfg_folder, file_prefix + nabu_settings.NABU_CONFIG_FILE_EXTENSION
        )
        return config, conf_file


class _ReconstructorMultiCor(_NabuBaseReconstructor):
    def __init__(
        self,
        nabu_config: dict,
        axis: Union[str, NabuPlane],
        cors: tuple,
        file_format,
        slice_index: Union[int, str] = "middle",
        output_file_prefix_pattern=None,
        *args,
        **kwargs,
    ):
        if not isinstance(cors, tuple):
            raise TypeError(
                f"cors are expected to be an instance of tuple. Get {type(cors)} instead"
            )
        self.__cors = cors
        self.__slice_index = slice_index
        self.__nabu_config = nabu_config
        self.file_format = file_format
        self._output_file_prefix_pattern = output_file_prefix_pattern

        super().__init__(*args, **kwargs)

    @property
    def cors(self) -> tuple:
        return self.__cors

    @property
    def slice_index(self) -> Union[int, str]:
        return self.__slice_index

    @property
    def nabu_config(self) -> dict:
        return self.__nabu_config

    def _process_config(
        self,
        config_to_dump: dict,
        config_file: str,
        file_format: str,
        info: Optional[str],
        process_name: str,
    ):
        """
        process provided configuration

        :param str info:
        """
        if self.dry_run is True or self.only_create_config_file():
            return ResultsRun(
                success=True,
                config=config_to_dump,
            )
        elif self.target is Target.LOCAL:
            _logger.info(f"run {info} for {self.scan} with {config_to_dump}")
            return self._run_nabu_multicor_locally(
                conf_file=config_file,
                file_format=file_format,
                config_to_dump=config_to_dump,
            )
        elif self.target is Target.SLURM:
            _logger.info(
                f"run {info} on slurm for {self.scan.path} with {config_to_dump}"
            )
            return self._run_nabu_multicor_on_slurm(
                conf_file=config_file,
                config_to_dump=config_to_dump,
                cluster_config=self.cluster_config.to_dict(),
                process_name=process_name,
                info=info,
            )
        else:
            raise ValueError(f"{self.target} is not recognized as a valid target")

    @staticmethod
    def convert_cor_from_rel_to_abs(scan, cor):
        if scan.dim_1 is not None:
            return cor + scan.dim_1 / 2.0
        else:
            _logger.warning("enable to get image half width. Set it to 1024")
            return cor + 1024

    @staticmethod
    def convert_cor_from_abs_to_rel(scan, cor):
        if scan.dim_1 is not None:
            return cor - scan.dim_1 / 2.0
        else:
            _logger.warning("enable to get image half width. Set it to 1024")
            return cor - 1024

    def _run_nabu_multicor_locally(
        self,
        conf_file: str,
        file_format: str,
        config_to_dump: dict,
    ) -> ResultsLocalRun:
        """
        run locally nabu for a single configuration file.

        :param str conf_file: path to the nabu .cfg file
        :param str file_format: format of the generated file
        :param dict config_to_dump: configuration saved in the .cfg as a dictionary
        :return: results of the local run
        :rtype: ResultsLocalRun
        """
        if not has_nabu:
            raise ImportError("Fail to import nabu")
        slice_index = slice_index_to_int(self.slice_index, scan=self.scan)

        cor_in_nabu_ref = tuple(
            [self.convert_cor_from_rel_to_abs(self.scan, cor) for cor in self.cors]
        )
        cor_in_nabu_ref = ",".join([str(cor) for cor in cor_in_nabu_ref])
        command = " ".join(
            (
                sys.executable,
                "-m",
                settings.NABU_MULTICOR_PATH,
                conf_file,  # input file
                f"{slice_index}",  # slice
                f"{cor_in_nabu_ref}",  # cor
            )
        )
        _logger.info(f'call nabu from "{command}"')

        self._process = subprocess.Popen(
            command,
            shell=True,
            cwd=self.scan.path,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        try:
            outs, errs = self._process.communicate()
        except (TimeoutError, KeyboardInterrupt):
            self._process.kill()
            outs, errs = self._process.communicate()

        file_prefix = get_nabu_multicor_file_prefix(self.scan)

        recons_vol_identifiers = utils.get_multi_cor_recons_volume_identifiers(
            slice_index=slice_index,
            location=config_to_dump["output"]["location"],
            file_prefix=file_prefix,
            scan=self.scan,
            file_format=file_format,
            cors=[
                self.convert_cor_from_rel_to_abs(self.scan, cor) for cor in self.cors
            ],
        )
        # convert back from abs ref to rel ref
        recons_vol_identifiers = {
            self.convert_cor_from_abs_to_rel(self.scan, cor): identifiers
            for cor, identifiers in recons_vol_identifiers.items()
        }
        return ResultsLocalRun(
            success=not nabu_std_err_has_error(errs),
            results_identifiers=recons_vol_identifiers.values(),
            std_out=outs,
            std_err=errs,
            config=config_to_dump,  # config_slices,
        )

    def _run_nabu_multicor_on_slurm(
        self,
        conf_file: str,
        config_to_dump: dict,
        cluster_config: dict,
        process_name: str,
        info: str,
    ) -> ResultSlurmRun:
        """
        Run a nabu reconstruction on slurm of a single configuration

        :return: results of the slurm run
        :rtype: ResultSlurmRun
        """
        if not isinstance(conf_file, str):
            raise TypeError(f"conf_file is expected to be a strg not {type(conf_file)}")
        if not isinstance(config_to_dump, dict):
            raise TypeError(
                f"config_to_dump is expected to be a strg not {type(config_to_dump)}"
            )
        if not is_slurm_available():
            raise RuntimeError("slurm not available")
        if not isinstance(cluster_config, dict):
            raise ValueError(
                f"cluster config is expected to be a dict not {type(cluster_config)}"
            )

        # create slurm cluster
        project_name = cluster_config.get(
            "job_name", "tomwer_{scan}_-_{process}_-_{info}"
        )
        project_name = project_name.format(
            scan=str(self.scan), process=process_name, info=info
        )
        # project name should not contain any spaces as it will be integrated in a script and interpreted.
        project_name = project_name.replace(" ", "_")
        cluster_config["job_name"] = project_name

        slice_index = slice_index_to_int(self.slice_index, scan=self.scan)
        cor_in_nabu_ref = tuple(
            [self.convert_cor_from_rel_to_abs(self.scan, cor) for cor in self.cors]
        )
        cor_in_nabu_ref = ",".join([str(cor) for cor in cor_in_nabu_ref])

        # submit job
        script_name = get_slurm_script_name(prefix="nabu")
        # for now force job name
        cluster_config["job_name"] = f"tomwer-nabu {conf_file}"
        job = SBatchScriptJob(
            slurm_config=cluster_config,
            script=(
                f"python3 -m {settings.NABU_MULTICOR_PATH} {conf_file} {slice_index} {cor_in_nabu_ref}",
            ),
            script_path=os.path.join(self.scan.path, "slurm_scripts", script_name),
            clean_script=False,
            working_directory=self.scan.working_directory,
        )
        future_slurm_job = submit_to_slurm_cluster(job)

        callbacks = self._get_futures_slurm_callback(config_to_dump)
        assert isinstance(
            callbacks, tuple
        ), f"callbacks is expected to an instance of tuple and not {type(callbacks)}"
        for callback in callbacks:
            future_slurm_job.add_done_callback(callback.process)

        return ResultSlurmRun(
            success=True,
            config=config_to_dump,
            future_slurm_jobs=(future_slurm_job,),
            std_out=None,
            std_err=None,
            job_id=job.job_id,
        )

    def preprocess_config(self, config):
        dataset_params = self.scan.get_nabu_dataset_info()
        if "dataset" in config:
            dataset_params.update(config["dataset"])
        config["dataset"] = dataset_params

        config["resources"] = utils.get_nabu_resources_desc(
            scan=self.scan, workers=1, method="local"
        )
        # force overwrite results
        if "output" not in config:
            config["output"] = {}
        config["output"].update({"overwrite_results": 1})

        cfg_folder = os.path.join(
            config["output"]["location"],
            nabu_settings.NABU_CFG_FILE_FOLDER,
        )
        os.makedirs(cfg_folder, exist_ok=True)

        cfg_folder = os.path.join(
            self.nabu_config["output"]["location"],
            nabu_settings.NABU_CFG_FILE_FOLDER,
        )
        os.makedirs(cfg_folder, exist_ok=True)

        conf_file = os.path.join(
            cfg_folder,
            f"{self.scan.get_dataset_basename()}_multi_cor"
            + nabu_settings.NABU_CONFIG_FILE_EXTENSION,
        )
        return config, conf_file

    def run(self) -> Iterable:
        nabu_config, conf_file = self.preprocess_config(deepcopy(self.nabu_config))

        # the policy is to save nabu .cfg file at the same location as the
        # force overwrite results

        # add some tomwer metadata and save the configuration
        # note: for now the section is ignored by nabu but shouldn't stay that way
        with utils.TomwerInfo(nabu_config) as config_to_dump:
            generate_nabu_configfile(
                conf_file,
                nabu_fullfield_default_config,
                config=config_to_dump,
                options_level="advanced",
            )

        results = self._process_config(
            config_to_dump=nabu_config,
            config_file=conf_file,
            file_format=self.file_format,
            info="nabu sa-axis reconstruction",
            process_name=self.process_name,
        )

        return results



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "19/07/2018"

import logging

from orangewidget import gui
from orangewidget.widget import Input, Output
from processview.core.manager import DatasetState
from silx.gui import qt

import tomwer.core.process.conditions.filters
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from tomwer.core.process.conditions import filters
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.conditions.filter import FileNameFilterWidget
from tomwer.utils import docstring

logger = logging.getLogger(__name__)


class NameFilterOW(SuperviseOW):
    name = "scan filter"
    id = "orange.widgets.tomwer.filterow"
    description = (
        "Simple widget which filter some data directory if the name"
        "doesn't match with the pattern defined."
    )
    icon = "icons/namefilter.svg"
    priority = 106
    keywords = ["tomography", "selection", "tomwer", "folder", "filter"]

    want_main_area = True
    resizing_enabled = True

    ewokstaskclass = tomwer.core.process.conditions.filters.FileNameFilterTask

    class Inputs:
        data = Input(name="data", type=TomwerScanBase, multiple=True)

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        """ """
        super().__init__(parent)

        self.widget = FileNameFilterWidget(parent=self)
        self.widget.setContentsMargins(0, 0, 0, 0)

        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self.widget)
        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        layout.addWidget(spacer)

    @Inputs.data
    def applyfilter(self, scan, *args, **kwargs):
        if scan is None:
            return
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(f"{scan} is expected to be an instance of {TomwerScanBase}")

        process = filters.FileNameFilterTask(
            inputs={
                "data": scan,
                "pattern": self.getPattern(),
                "filter_type": self.getActiveFilter(),
                "invert_result": self.invertFilterAction(),
                "serialize_output_data": False,  # avoid spending time on scan serialization / deserialization when use orange
            }
        )
        process.run()
        out = process.outputs.data
        if out is not None:
            self.set_dataset_state(dataset=scan, state=DatasetState.SUCCEED)
            logger.processSucceed(f"{scan} pass through filter")
            self._signalScanReady(scan)
        else:
            self.set_dataset_state(dataset=scan, state=DatasetState.FAILED)
            logger.processFailed(f"{scan} NOT pass through filter")

    @docstring(SuperviseOW)
    def reprocess(self, dataset):
        self.applyfilter(dataset)

    def _signalScanReady(self, scan):
        self.Outputs.data.send(scan)

    def getPattern(self):
        return self.widget.getPattern()

    def setPattern(self, pattern):
        self.widget.setPattern(pattern)

    def getActiveFilter(self):
        return self.widget.getActiveFilter()

    def setActiveFilter(self, filter):
        self.widget.setActiveFilter(filter)

    def invertFilterAction(self):
        return self.widget.invertFilterAction()



import logging
import weakref
import numpy
from typing import Any, Union, Optional
from contextlib import AbstractContextManager

from silx.gui import qt
from silx.gui.plot import PlotWindow
from nxtomo.nxobject.nxdetector import ImageKey
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.tomwer_object import TomwerObject
from tomwer.gui.stitching.metadataholder import QStitchingMetadata
from tomwer.gui import icons as tomwer_icons

_logger = logging.getLogger(__name__)


class RawStitchingPlot(PlotWindow):
    """
    Plot displaying the different TomwerObject (volume or scan) at the specify position.
    This widget is used to help users find initial positions for stitching over anspecific axis.

    It will display a single slice or projection.

    The design is that it will keep up to date with TomwerObject if the widget is "active".
    Else it will wait for either a manual update (calling updateImages) or to be activated.

    tomo objects are weakref and stitching metadata will always be connected. But ignore if not activated.
    Activation mecanism is here because this can be memory consuming and we want by default to avoid this mecanism

    Warning: for now the interface expects to have frames homogeneous space. So if a scan / volume gets invertions (y downward)... we expect all
    scans / volumes to have the same. This is the same at nabu level.
    """

    class ActivateContext(AbstractContextManager):
        """
        simple context to turn off : turn on image update
        """

        def __init__(self, rawStitchingPlot, activate) -> None:
            super().__init__()
            self._rawStitchingPlot = rawStitchingPlot
            self._activate = activate

        def __enter__(self) -> Any:
            self._activateStatus = self._rawStitchingPlot.isActive()
            self._rawStitchingPlot.setActive(self._activate)

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._rawStitchingPlot.setActive(self._activateStatus)

    def __init__(self, parent=None, alpha_values=False, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self._activated = True
        self._tomo_objs = dict({})
        self._slice_for_preview = "middle"
        self._scan_reading_order = {}
        # for each scan store if the rotation angles are store in increasing order or decreasing.
        # this way we can try to know if the order is inverted between different scans.
        self._flipudFrames = False
        # should we flip up / down frame (once management of metadata is done). For scan we can get fliplr and flip ud information already
        # but not for volumes
        self._fliplrFrames = False
        self._firstDisplay = True
        # bool to know if we are currently displaying the BW mode
        self._blackAndWhiteCompositeMode = False
        # result of the B&W composite mode
        self._BWImage = None
        self._compositeOffset = None, None

        # add alpha values widget
        if alpha_values:
            self._alphaValuesWidget = AlphaValuesTableWidget(self, plot=self)
            self._alphaValuesWidgetDockWidget = qt.QDockWidget(self)
            self._alphaValuesWidgetDockWidget.setWidget(self._alphaValuesWidget)
            self.addDockWidget(
                qt.Qt.BottomDockWidgetArea, self._alphaValuesWidgetDockWidget
            )
        else:
            self._alphaValuesWidget = None

        # composite  B&W mode
        self._modeToolbar = qt.QToolBar("display mode", self)

        icon = tomwer_icons.getQIcon("compare_mode_a_minus_b")
        self._substractAction = qt.QAction(icon, "Subtract in BW mode", self)
        self._substractAction.setCheckable(True)
        self._substractAction.setChecked(False)
        self._substractAction.toggled.connect(self.__displayModeChanged)
        self._modeToolbar.addAction(self._substractAction)
        self._modeToolbar.setVisible(
            False
        )  # for now this mode is a proto. Complex to do to keep some performances.

        self.addToolBar(self._modeToolbar)

    def __displayModeChanged(self):
        self._blackAndWhiteCompositeMode = self._substractAction.isChecked()
        self._updateImages()

    def setFlipLRFrames(self, flip: bool) -> None:
        if flip != self._fliplrFrames:
            self._fliplrFrames = flip
            self._updateImages()

    def setFlipUDFrames(self, flip: bool):
        if flip != self._flipudFrames:
            self._flipudFrames = flip
            self._updateImages()

    def setSliceForPreview(self, slice_for_preview: Union[str, int]):
        if self._slice_for_preview != slice_for_preview:
            self._slice_for_preview = slice_for_preview
            self._updateImages()

    def setActive(self, active: bool):
        if active == self._activated:
            # avoid any update if not necessary
            return
        self._activated = active
        if active:
            # if has been activated update plot to be up to date
            self._updateImages()

    def isActive(self) -> bool:
        return self._activated

    @staticmethod
    def getMinAxisPosition(tomo_obj: Optional[TomwerObject], axis):
        if tomo_obj is None or tomo_obj.stitching_metadata is None:
            return 0
        else:
            return tomo_obj.stitching_metadata.get_abs_position_px(axis=axis) or 0

    @staticmethod
    def getRotationAngleDirection(scan: TomwerScanBase):
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"scan is expected to be an instance of {TomwerScanBase} instead of {type(scan)}"
            )

        rotation_angle_ref = numpy.array(scan.rotation_angle)
        return rotation_angle_ref[
            numpy.asarray(scan.image_key_control) == ImageKey.PROJECTION.value
        ]

    def addTomoObj(self, tomo_obj: TomwerObject):
        if not isinstance(tomo_obj, TomwerObject):
            raise TypeError(
                f"tomo_obj is expected to be an instance of {TomwerObject}. {type(tomo_obj)} provided"
            )
        elif tomo_obj.stitching_metadata is None:
            raise ValueError(
                "tomo obj is expected to have a stitching_metadata defined yet"
            )

        if len(self._tomo_objs) > 0:
            first_tomo_obj_id = tuple(self._tomo_objs.keys())[0]
            first_tomo_obj = self._tomo_objs[first_tomo_obj_id]()
        else:
            first_tomo_obj_id = None
            first_tomo_obj = None

        if not isinstance(first_tomo_obj, (TomwerScanBase, type(None))) and isinstance(
            tomo_obj, TomwerScanBase
        ):
            # We cannot display both scan and volumes. Especially because we need to define reading order for scan
            # and we need to keep a reference for scans (first one over the stitching axis)
            _logger.error(
                f"get instances of both {TomwerScanBase} and {TomwerVolumeBase}. Case not handled. Skip it"
            )
            return

        self._tomo_objs[tomo_obj.get_identifier().to_str()] = weakref.ref(tomo_obj)
        if isinstance(tomo_obj.stitching_metadata, QStitchingMetadata):
            tomo_obj.stitching_metadata.sigChanged.connect(
                self._tomoObjMetadataHasChanged
            )

        # reorder tomo_objs to keep it axis 0 ordered
        self._tomo_objs = dict(
            sorted(
                self._tomo_objs.items(),
                key=lambda item: self.getMinAxisPosition(item[1](), axis=0),
                reverse=True,  # scan are expected to be orderred alonx axis 0 and decreasing
            )
        )

        self._alphaValuesWidget.addTomoObj(tomo_obj)
        if isinstance(tomo_obj, TomwerScanBase):
            self._scan_reading_order[tomo_obj.get_identifier().to_str()] = (
                self.getRotationAngleDirection(tomo_obj)
            )
            if first_tomo_obj_id != tuple(self._tomo_objs.keys())[0]:
                # if the order of scans might have change we need to update the full stack (to handle reading order)
                self._updateImages()
                return

        self._updateImage(tomo_obj)

    def _tomoObjMetadataHasChanged(self, *args, **kwargs):
        sender = self.sender()
        if not isinstance(sender, QStitchingMetadata):
            _logger.error(
                "_tomoObjMetadataHasChanged is expected to be conencted with QStitchingMetadata only"
            )
        else:
            tomo_obj = sender.tomo_obj
            if tomo_obj is not None:
                self._updateImage(tomo_obj=tomo_obj)

    def removeTomoObj(self, tomo_obj: TomwerObject):
        if not isinstance(tomo_obj, TomwerObject):
            raise TypeError(
                f"tomo_obj is expected to be an instance of {TomwerObject}. {type(tomo_obj)} provided"
            )

        self._tomo_objs.pop(tomo_obj.get_identifier().to_str(), None)
        self._scan_reading_order.pop(tomo_obj.get_identifier().to_str(), None)

        # reorder tomo_objs to keep it axis 0 ordered
        self._tomo_objs = dict(
            sorted(
                self._tomo_objs.items(),
                key=lambda item: self.getMinAxisPosition(item[1](), axis=0),
                reverse=True,
            )
        )
        if len(self._tomo_objs) > 0:
            first_tomo_obj_id = tuple(self._tomo_objs.keys())[0]
            first_tomo_obj = self._tomo_objs[first_tomo_obj_id]()
        else:
            first_tomo_obj_id = None
            first_tomo_obj = None

        if first_tomo_obj is tomo_obj and isinstance(first_tomo_obj, TomwerScanBase):
            # if a scan has been added and if he is the first one over the axis. Then the reading order was the 'ref order'
            # and we need to update all images
            self._updateImages()
        else:
            # else we can simple remove the image it will not affect others
            self.removeImage(tomo_obj.get_identifier().to_str())

    def _updateImage(self, tomo_obj: TomwerObject, recompute_composite=True):
        if not self._activated:
            return

        if not isinstance(tomo_obj, TomwerObject):
            raise TypeError(
                f"tomo_obj is expected to be an instance of {TomwerObject}. {type(tomo_obj)} provided"
            )

        if recompute_composite:
            self._recomputeBlackAndWhiteCompositeImage()

        frame = self.getFrame(tomo_obj=tomo_obj)

        if frame is None:
            return
        frame_origin = (
            tomo_obj.stitching_metadata.get_abs_position_px(axis=2) or 0,
            tomo_obj.stitching_metadata.get_abs_position_px(axis=0) or 0,
        )
        # replace frame by the composite if needed
        if self._blackAndWhiteCompositeMode:
            frame = self._BWImage[
                self._compositeOffset[0] + frame_origin[1] : frame.shape[0],
                self._compositeOffset[1] + frame_origin[0] : frame.shape[1],
            ]

        identifier = tomo_obj.get_identifier().to_str()
        self.addImage(
            data=frame[
                ::-1
            ],  # for coherence with axis and keep frame ordered we need to invert this one
            legend=identifier,
            origin=frame_origin,
            resetzoom=self._firstDisplay,
        )
        self._firstDisplay = False
        img = self.getImage(identifier)
        img.setAlpha(0.5)  # TODO: look to be able to tune this value from interface...

    def clearTomoObjs(self):
        tomo_objs = tuple(self._tomo_objs.values())
        for tomo_obj in tomo_objs:
            self.removeTomoObj(tomo_obj())
        self._alphaValuesWidget.clearTomoObjs()

    def setTomoObjs(self, tomo_objs):
        self.clearTomoObjs()
        # avoid to update images each time
        with RawStitchingPlot.ActivateContext(self, activate=False):
            for tomo_obj in tomo_objs:
                self.addTomoObj(tomo_obj=tomo_obj)
        self._alphaValuesWidget.setTomoObjs(tomo_objs)
        self._updateImages()

    def _getReadingOrderRef(self):
        return self._scan_reading_order[tuple(self._tomo_objs.keys())[0]]

    def getFrame(self, tomo_obj):
        if isinstance(tomo_obj, TomwerVolumeBase):
            try:
                frame = tomo_obj.get_slice(index=self._slice_for_preview, axis=1)
            except IndexError:
                _logger.error(
                    f"requested slice doesn't exist (index=={self._slice_for_preview}, axis==1)"
                )
                return None

            if self._flipudFrames:
                frame = numpy.flipud(frame)
            if self._fliplrFrames:
                frame = numpy.fliplr(frame)
            return frame

        elif isinstance(tomo_obj, TomwerScanBase):
            scan = tomo_obj

            # compute flat field
            # TODO: this must be done in another thread... but sync with update might be difficult...
            try:
                reduced_darks, reduced_darks_infos = scan.load_reduced_darks(
                    return_info=True
                )
            except Exception:
                _logger.warning(
                    f"no reduced dark found for {scan}. Please compute them to get a better display."
                )
            else:
                scan.set_reduced_darks(reduced_darks, darks_infos=reduced_darks_infos)

            try:
                reduced_flats, reduced_flats_info = scan.load_reduced_flats(
                    return_info=True
                )
            except Exception:
                _logger.warning(
                    f"no reduced flat(s) found for {scan}. Please compute them to get a better display."
                )
            else:
                scan.set_reduced_flats(reduced_flats, flats_infos=reduced_flats_info)

            reading_order_ref = self._getReadingOrderRef()
            reading_order = self._scan_reading_order[scan.get_identifier().to_str()]

            invert_reading_order = not numpy.allclose(
                reading_order_ref, reading_order, atol=10e-1
            )

            slice_to_use = self._slice_for_preview
            if slice_to_use == "first":
                slice_to_use = 0
            elif slice_to_use == "last":
                slice_to_use = -1
            elif slice_to_use == "middle":
                slice_to_use = len(scan.projections) // 2

            proj_idx = sorted(scan.projections.keys(), reverse=invert_reading_order)[
                slice_to_use
            ]
            ff = scan.flat_field_correction(
                (scan.projections[proj_idx],),
                (slice_to_use,),
            )[0]

            if scan.get_x_flipped(default=False) ^ self._fliplrFrames:
                ff = numpy.fliplr(ff)
            if scan.get_y_flipped(default=False) ^ self._flipudFrames:
                ff = numpy.flipud(ff)

            return ff
        else:
            raise TypeError(f"tomo_obj of type {type(tomo_obj)} is not handled")

    def _recomputeBlackAndWhiteCompositeImage(self):
        def get_min_max(tomo_obj, axis):
            axis_pos = tomo_obj.stitching_metadata.get_abs_position_px(axis=axis) or 0
            frame_axis = 0 if axis == 0 else 1
            return (
                axis_pos,
                axis_pos + self.getFrame(tomo_obj=tomo_obj).shape[frame_axis],
            )

        import sys
        from math import ceil

        min_x = sys.float_info.max
        min_y = sys.float_info.max
        max_x = sys.float_info.min
        max_y = sys.float_info.min
        for tomo_obj in self._tomo_objs.values():
            if tomo_obj() is None:
                continue
            l_min_x, l_max_x = get_min_max(tomo_obj(), axis=2)
            l_min_y, l_max_y = get_min_max(tomo_obj(), axis=0)
            min_x = min(min_x, l_min_x)
            max_x = max(max_x, l_max_x)
            min_y = min(min_y, l_min_y)
            max_y = max(max_y, l_max_y)
        self._compositeOffset = (min_y, min_x)
        self._BWImage = numpy.zeros(
            shape=(
                ceil(max_y - min_y),
                ceil(max_x - min_x),
            ),
            dtype=numpy.float32,
        )
        # update BW image
        for tomo_obj in self._tomo_objs.values():
            if tomo_obj() is None:
                continue
            else:
                frame = self.getFrame(tomo_obj=tomo_obj())
                position_y = (
                    tomo_obj().stitching_metadata.get_abs_position_px(axis=0) or 0
                )
                position_x = (
                    tomo_obj().stitching_metadata.get_abs_position_px(axis=2) or 0
                )
                self._BWImage[
                    position_y - min_y : frame.shape[0] + position_y - min_y,
                    position_x - min_x : position_x + frame.shape[1] - min_x,
                ] -= frame

    def _updateImages(self):
        self.clearImages()
        first_display = self._firstDisplay
        if self._blackAndWhiteCompositeMode:
            self._recomputeBlackAndWhiteCompositeImage()
        for tomo_obj in self._tomo_objs.values():
            if tomo_obj() is None:
                continue
            else:
                self._updateImage(
                    tomo_obj=tomo_obj(),
                    recompute_composite=self._blackAndWhiteCompositeMode,
                )
        if first_display and len(self._tomo_objs) > 0 and self._activated:
            # work around to handle first plotting when just activated and can get several objects
            self.resetZoom()
            self._firstDisplay = False


class AlphaValuesTableWidget(qt.QTableWidget):
    """
    Widget to define alpha values of a list of tomo object.
    Meant to work with the RawPlotWidget used for stitching
    """

    COLUMNS = "legend", "alpha value"

    def __init__(self, parent=None, plot=None):
        super().__init__(parent)
        if plot is not None:
            self.__plot = weakref.ref(plot)
        self._sliders = {}

    def getPlot(self):
        if self.__plot is None:
            return None
        else:
            return self.__plot()

    def clear(self):
        for slider, label in self._sliders.items():
            slider.valueChanged.disconnect(self._alphaValueChanged)

        self._sliders.clear()
        super().clear()

    def addTomoObj(self, tomo_obj: TomwerObject):
        n_objs = self.rowCount() + 1
        self.setRowCount(n_objs)
        self.setColumnCount(len(self.COLUMNS))
        self.setHorizontalHeaderLabels(self.COLUMNS)
        self.horizontalHeader().setSectionResizeMode(qt.QHeaderView.ResizeToContents)
        self.verticalHeader().hide()

        _item_legend = qt.QTableWidgetItem()
        plot_legend = tomo_obj.get_identifier().to_str()
        short_descrition = tomo_obj.get_identifier().short_description()
        _item_legend.setText(short_descrition)
        self.setItem(n_objs - 1, 0, _item_legend)

        slider = AlphaValueSlider(self)
        slider.valueChanged.connect(self._alphaValueChanged)
        self.setCellWidget(n_objs - 1, 1, slider)
        self._sliders[slider] = (plot_legend, short_descrition)

    def clearTomoObjs(self):
        # disconnect sliders
        for slider in self._sliders.keys():
            slider.valueChanged.disconnect(self._alphaValueChanged)
        self._sliders.clear()
        self.setRowCount(len(self._sliders))
        self.setColumnCount(len(self.COLUMNS))
        self.setHorizontalHeaderLabels(self.COLUMNS)
        self.horizontalHeader().setSectionResizeMode(qt.QHeaderView.ResizeToContents)
        self.verticalHeader().hide()

    def setTomoObjs(self, tomo_objs: tuple):
        self.clearTomoObjs()
        self.clear()

        for tomo_obj in tomo_objs:
            self.addTomoObj(tomo_obj=tomo_obj)

    def _alphaValueChanged(self, *args, **kwargs):
        sender = self.sender()  # pylint: disable=E1101
        assert isinstance(
            sender, AlphaValueSlider
        ), f"sender is expected to be an instance of AlphaValueSlider. get {type(sender)}"
        legend = self._sliders.get(sender, (None, None))[0]
        if legend is not None:
            plot = self.getPlot()
            if plot is None:
                return
            else:
                image_item = plot.getImage(legend)
                if image_item is not None:
                    image_item.setAlpha(
                        sender.value() / 255
                    )  # silx define value in [0, 1.0] when AlphaValueSlider in [0, 255]


class AlphaValueSlider(qt.QTableWidget):
    """
    slider to define alpha value with a slider and top button on each side to quickly set values to boundaries
    """

    valueChanged = qt.Signal(int)
    " emit when the slider value changed"

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setLayout(qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        # `invisible` button
        invisibleIcon = tomwer_icons.getQIcon("invisible")
        self._invisibleButton = qt.QPushButton(invisibleIcon, "", self)
        self._invisibleButton.setFlat(True)
        self.layout().addWidget(self._invisibleButton)
        # `slider`
        self._slider = qt.QSlider(qt.Qt.Horizontal, self)
        self._slider.setRange(0, 255)
        self._slider.setValue(255)
        self.layout().addWidget(self._slider)

        # `visible` button
        visibleIcon = tomwer_icons.getQIcon("visible")
        self._visibleButton = qt.QPushButton(visibleIcon, "", self)
        self._visibleButton.setFlat(True)
        self.layout().addWidget(self._visibleButton)

        # connect signal / slot
        self._invisibleButton.released.connect(self._setInvisible)
        self._visibleButton.released.connect(self._setVisible)

        # expose API
        self._slider.valueChanged.connect(self.valueChanged)

    def _setInvisible(self, *args, **kwargs):
        self._slider.setValue(0)

    def _setVisible(self, *args, **kwargs):
        self._slider.setValue(255)

    def value(self) -> int:
        return self._slider.value()

    def setValue(self, value: int):
        self._slider.setValue(int(value))



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/05/2023"


import pytest
from tomwer.core.volume.hdf5volume import HDF5Volume
from tomwer.gui.visualization.reconstructionparameters import ReconstructionParameters

from tomwer.tests.conftest import qtapp  # noqa F401


@pytest.mark.parametrize("phase_method", ("", "CTF", "Paganin"))
def test_ReconstructionParameters(qtapp, phase_method):  # noqa F401
    window = ReconstructionParameters()
    volume = HDF5Volume(
        file_path="test.hdf5",
        data_path="data",
        data=None,
        metadata={
            "nabu_config": {
                "reconstruction": {
                    "method": "FBP",
                },
                "phase": {
                    "method": phase_method,
                    "delta_beta": 110.0,
                },
            },
            "processing_options": {
                "reconstruction": {
                    "voxel_size_cm": (0.2, 0.2, 0.2),
                    "rotation_axis_position": 104,
                    "enable_halftomo": True,
                    "fbp_filter_type": "Hilbert",
                    "sample_detector_dist": 0.4,
                },
                "take_log": {
                    "log_min_clip": 1.0,
                    "log_max_clip": 10.0,
                },
            },
        },
    )
    window.setVolumeMetadata(metadata=volume.metadata)

    assert window._methodQLE.text() == "FBP"
    assert window._paganinQLE.text() == phase_method
    assert window._deltaBetaQLE.text() == "110.0"
    assert window._distanceQLE.text() == "0.4"
    assert window._pixelSizeQLE.text() == "0.2"
    assert window._corQLE.text() == "104.00"
    assert window._halfTomoCB.isChecked()
    assert window._fbpFilterQLE.text() == "Hilbert"
    assert window._minLogClipQLE.text() == "1.0"
    assert window._maxLogClipQLE.text() == "10.0"

    window.setVolumeMetadata(metadata=None)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/
"""
data watcher classes used to define the status of an acquisition for EDF
acquisitions
"""

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "30/09/2019"

import logging

import h5py

from .datawatcherprocess import _DataWatcherProcess

_logger = logging.getLogger(__name__)


class _DataWatcherProcessHDF5(_DataWatcherProcess):
    """
    look for hdf5 information
    """

    def __init__(self, dataDir, srcPattern, destPattern):
        super(_DataWatcherProcessHDF5, self).__init__(dataDir, srcPattern, destPattern)
        if h5py.is_hdf5(dataDir):
            self._nxtomo_file = dataDir

    def _removeAcquisition(self, scanID, reason):
        _logger.warning(
            "removing acquisition is not done for hdf5 data watcher " "process"
        )

    def is_data_complete(self):
        # For now data complete and is abort_is not handled for hdf5 and Nxtomo
        return True

    def is_abort(self):
        # For now data complete and is abort_is not handled for hdf5 and Nxtomo
        return False


class _BlissScanWatcherProcess(_DataWatcherProcess):
    def __init__(self, dataDir, srcPattern=None, destPattern=None):
        super().__init__(dataDir, srcPattern, destPattern)
        if h5py.is_hdf5(dataDir):
            self._blissScanFile = dataDir

    def _removeAcquisition(self, scanID, reason):
        _logger.warning(
            "remoing acquisition is not done for hdf5 data watcher " "process"
        )

    def is_data_complete(self):
        # For now data complete and is abort_is not handled for hdf5 and Nxtomo
        return True

    def is_abort(self):
        # For now data complete and is abort_is not handled for hdf5 and Nxtomo
        return False



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/04/2019"


import os
import shutil
import tempfile
import unittest

import h5py
from tomoscan.io import get_swmr_mode
import numpy
from silx.io.utils import h5py_read_dataset

from tomwer.core.process.reconstruction.axis.mode import AxisMode
from tomwer.core.process.reconstruction.axis.params import AxisRP
from tomwer.core.process.task import Task
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.scanutils import MockEDF, MockNXtomo
from tomwer.tests.datasets import TomwerCIDatasets

from ..reconstruction.axis.axis import AxisTask


class TestAxisIO(unittest.TestCase):
    """Test inputs and outputs types of the handler functions"""

    @staticmethod
    def _random_calc(scan):
        return numpy.random.random()

    def setUp(self):
        self.scan_folder = tempfile.mkdtemp()

        self.scan = MockEDF.mockScan(
            scanID=self.scan_folder, nRadio=10, nRecons=1, nPagRecons=4, dim=10
        )
        self.recons_params = AxisRP()

        # set the axis url to be used
        projections = self.scan.projections
        urls = list(projections.values())
        self.scan.axis_params = AxisRP()
        self.scan.axis_params.axis_url_1 = urls[0]
        self.scan.axis_params.axis_url_2 = urls[1]

    def tearDown(self):
        shutil.rmtree(self.scan_folder)

    def testInputOutput(self):
        """Test that io using TomoBase instance work"""
        for input_type in (dict, TomwerScanBase):
            for serialize_output_data in (True, False):
                with self.subTest(
                    serialize_output_data=serialize_output_data,
                    input_type=input_type,
                ):
                    input_obj = self.scan
                    if input_obj is dict:
                        input_obj = input_obj.to_dict()
                    axis_process = AxisTask(
                        inputs={
                            "axis_params": self.recons_params,
                            "data": input_obj,
                            "serialize_output_data": serialize_output_data,
                        }
                    )

                    # patch the axis process
                    axis_process._CALCULATIONS_METHODS[AxisMode.centered] = (
                        TestAxisIO._random_calc
                    )

                    axis_process.run()
                    out = axis_process.outputs.data
                    if serialize_output_data:
                        self.assertTrue(isinstance(out, dict))
                    else:
                        self.assertTrue(isinstance(out, TomwerScanBase))


class TestAxis(unittest.TestCase):
    """Test the axis process"""

    def setUp(self):
        self.recons_params = AxisRP()
        self.recons_params.mode = "centered"

    def test_process_saved_edf(self):
        """Test that if process is called, the tomwer.h5 file is created
        and is correctly saving information regarding the center of position
        """
        self.tempdir = tempfile.mkdtemp()
        mock = MockEDF(scan_path=self.tempdir, n_radio=10, n_ini_radio=10)
        scan = EDFTomoScan(mock.scan_path)
        self.recons_params.mode = "centered"
        axis_process = AxisTask(
            inputs={
                "data": scan,
                "axis_params": self.recons_params,
                "serialize_output_data": False,
            }
        )

        axis_process.run()
        self.assertTrue(os.path.exists(scan.process_file))

        with h5py.File(scan.process_file, "r", swmr=get_swmr_mode()) as h5f:
            self.assertTrue("entry" in h5f)
            self.assertTrue("tomwer_process_0" in h5f["entry"])
            group_axis = h5f["entry"]["tomwer_process_0"]
            self.assertTrue("configuration" in group_axis)
            self.assertTrue("program" in group_axis)
            self.assertTrue("results" in group_axis)
            self.assertTrue("center_of_rotation" in group_axis["results"])
            axis_value = h5py_read_dataset(group_axis["results"]["center_of_rotation"])

        processes = Task.get_processes_frm_type(
            process_file=scan.process_file, process_type=AxisTask, entry="entry"
        )
        self.assertEqual(len(processes), 1)
        self.assertEqual(processes[0].results["center_of_rotation"], axis_value)

        # make sure the process file can be loaded by an instance of AxisRP
        new_axis_params = AxisRP()
        new_axis_params.set_relative_value(-10000)
        new_axis_params.set_position_frm_par_file(
            file_path=scan.process_file, entry=None
        )
        self.assertEqual(
            new_axis_params.relative_cor_value, scan.axis_params.relative_cor_value
        )

    def test_process_saved_hdf5(self):
        """Test that if process is called, the tomwer.h5 file is created
        and is correctly saving information regarding the center of position
        """
        self.tempdir = tempfile.mkdtemp()
        dim = 10
        mock = MockNXtomo(
            scan_path=self.tempdir, n_proj=10, n_ini_proj=10, scan_range=180, dim=dim
        )
        mock.add_alignment_radio(index=10, angle=90)
        mock.add_alignment_radio(index=10, angle=0)
        scan = mock.scan
        self.recons_params.mode = "centered"

        # check data url take
        axis_process = AxisTask(
            inputs={
                "data": scan,
                "axis_params": self.recons_params,
                "serialize_output_data": False,
            }
        )
        axis_process.run()
        # make sure center of position has been computed
        self.assertTrue(os.path.exists(scan.process_file))

        with h5py.File(scan.process_file, "r", swmr=get_swmr_mode()) as h5f:
            self.assertTrue("entry" in h5f)
            self.assertTrue("tomwer_process_0" in h5f["entry"])
            group_axis = h5f["entry"]["tomwer_process_0"]
            self.assertTrue("configuration" in group_axis)
            self.assertTrue("program" in group_axis)
            self.assertTrue("results" in group_axis)
            self.assertTrue("center_of_rotation" in group_axis["results"])
            axis_value = h5py_read_dataset(group_axis["results"]["center_of_rotation"])
        self.assertTrue(-dim / 2 <= axis_value <= dim / 2)
        processes = Task.get_processes_frm_type(
            process_file=scan.process_file, process_type=AxisTask
        )
        self.assertEqual(len(processes), 1)
        self.assertEqual(processes[0].results["center_of_rotation"], axis_value)

        # make sure the process file can be loaded by an instance of AxisRP
        new_axis_params = AxisRP()
        new_axis_params.set_relative_value(-10000)
        new_axis_params.set_position_frm_par_file(file_path=scan.process_file)
        self.assertEqual(
            new_axis_params.relative_cor_value, scan.axis_params.relative_cor_value
        )


class TestAxisRP(unittest.TestCase):
    """Test the class used for AxisProcess configuration"""

    def setUp(self):
        self.axis_rp = AxisRP()
        self.tmp_folder = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmp_folder)

    def testSetPositionFromParFile(self):
        """Test that the position can be managed from a par file"""
        old_value = self.axis_rp.value_ref_tomwer

        with self.assertRaises(ValueError):
            self.axis_rp.set_position_frm_par_file("ghost_file.par", force=False)
        self.assertEqual(self.axis_rp.value_ref_tomwer, old_value)

        existing_empty_file = os.path.join(self.tmp_folder, "par_file.par")
        with open(existing_empty_file, "w") as _file:
            _file.write("toto")

        with self.assertRaises(KeyError):
            self.axis_rp.set_position_frm_par_file(existing_empty_file, force=False)
        self.assertEqual(self.axis_rp.value_ref_tomwer, old_value)

        valid_par_file = TomwerCIDatasets.get_dataset(
            "edf_datasets/scan_3_/scan_3_slice.par"
        )
        assert os.path.isfile(valid_par_file)
        self.axis_rp.set_position_frm_par_file(valid_par_file, force=False)
        self.assertEqual(self.axis_rp.value_ref_tomwer, 1024.742042)

        self.axis_rp.set_position_frm_par_file(existing_empty_file, force=True)
        self.assertEqual(self.axis_rp.value_ref_tomwer, None)

        self.axis_rp.set_position_frm_par_file("ghost_file.par", force=True)
        self.assertEqual(self.axis_rp.value_ref_tomwer, None)

        self.axis_rp.set_position_frm_par_file(valid_par_file, force=True)
        self.assertEqual(self.axis_rp.value_ref_tomwer, 1024.742042)


class TestSinogramAlgorithm(unittest.TestCase):
    def setUp(self) -> None:
        self.tmp_folder = tempfile.mkdtemp()
        dim = 512
        self.scan = MockNXtomo(
            scan_path=self.tmp_folder,
            n_proj=10,
            n_ini_proj=10,
            create_ini_dark=True,
            create_final_flat=True,
            scan_range=360,
            dim=dim,
        ).scan

    def tearDown(self):
        shutil.rmtree(self.tmp_folder)

    def test_growing_window_sinogram(self):
        recons_params = AxisRP()
        axis_process = AxisTask(
            inputs={"axis_params": recons_params, "data": self.scan}
        )
        recons_params.mode = AxisMode.growing_window_sinogram
        recons_params.use_sinogram = True
        recons_params.sinogram_subsampling = 1
        axis_process.run()

    def test_sliding_window(self):
        recons_params = AxisRP()
        axis_process = AxisTask(
            inputs={
                "axis_params": recons_params,
                "data": self.scan,
                "serialize_output_data": False,
            }
        )
        recons_params.mode = AxisMode.sliding_window_sinogram
        recons_params.side = "left"
        recons_params.sinogram_subsampling = 1
        recons_params.use_sinogram = True
        axis_process.run()

    def test_sino_coarse_to_fine(self):
        recons_params = AxisRP()
        axis_process = AxisTask(
            inputs={
                "axis_params": recons_params,
                "data": self.scan,
                "serialize_output_data": False,
            }
        )
        recons_params.mode = AxisMode.sino_coarse_to_fine
        recons_params.sinogram_subsampling = 1
        recons_params.sinogram_line = 2
        axis_process.run()



# # coding: utf-8
# # /*##########################################################################
# #
# # Copyright (c) 2016-2017 European Synchrotron Radiation Facility
# #
# # Permission is hereby granted, free of charge, to any person obtaining a copy
# # of this software and associated documentation files (the "Software"), to deal
# # in the Software without restriction, including without limitation the rights
# # to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# # copies of the Software, and to permit persons to whom the Software is
# # furnished to do so, subject to the following conditions:
# #
# # The above copyright notice and this permission notice shall be included in
# # all copies or substantial portions of the Software.
# #
# # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# # IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# # THE SOFTWARE.
# #
# # ###########################################################################*/

# __authors__ = ["H. Payno"]
# __license__ = "MIT"
# __date__ = "16/11/2020"


# import logging

# from orangewidget import gui
# from orangewidget.settings import Setting
# from orangewidget.widget import Input, Output
# from processview.core.manager import DatasetState, ProcessManager
# from silx.gui import qt

# import tomwer.core.process.control.volumesymlink
# from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
# from tomwer.core.process.control.volumesymlink import (
#     OutputType,
#     VolumeSymbolicLinkTask,
# )
# from tomwer.core.scan.scanbase import TomwerScanBase

# _logger = logging.getLogger(__name__)


# class VolumeSymLinkOW(SuperviseOW):
#     """
#     link volume reconstructed at proposal file if possible else under
#     the given folder
#     """

#     name = "volume symbolic link"
#     id = "orange.widgets.tomwer.control.VolumeSymLinkOW.VolumeSymLinkOW"
#     description = (
#         "create a symbolic link for the latest volume reconstructed on a "
#         "given folder"
#     )
#     icon = "icons/volumesymlink.svg"
#     priority = 200
#     keywords = [
#         "hdf5",
#         "tomwer",
#         "folder",
#         "volume",
#         "symlink",
#         "tomography",
#         "symbolic link",
#     ]

#     want_main_area = True
#     resizing_enabled = True

#     ewokstaskclass = tomwer.core.process.control.volumesymlink.VolumeSymbolicLinkTask

#     _scans = Setting(list())

#     class Inputs:
#         data = Input(name="data", type=TomwerScanBase, doc="scan to be process")

#     class Outputs:
#         data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

#     def __init__(self, parent=None):
#         super().__init__(parent)
#         self._mainWidget = _OutputDir(parent=self)
#         layout = gui.vBox(self.mainArea, self.name).layout()
#         layout.addWidget(self._mainWidget)

#     @Inputs.data
#     def process(self, scan):
#         if scan is None:
#             return
#         symlink_process = VolumeSymbolicLinkTask(
#             inputs={
#                 "data": scan,
#                 "output_type": self._mainWidget.getOutputType(),
#                 "output_folder": self._mainWidget.getOutputFolder(),
#                 "serialize_output_data": False,
#             }
#         )
#         ProcessManager().notify_dataset_state(
#             dataset=self.scan,
#             process=self,
#             state=DatasetState.ON_GOING,
#         )
#         try:
#             symlink_process.run()
#         except Exception:
#             mess = f"Failed to create symbolic link for {scan}"
#             _logger.processFailed(mess)
#             ProcessManager().notify_dataset_state(
#                 dataset=self.scan, process=self, state=DatasetState.FAILED, details=mess
#             )
#         else:
#             mess = f"Create symbolic link for {scan}"
#             _logger.processSucceed(mess)
#             ProcessManager().notify_dataset_state(
#                 dataset=self.scan,
#                 process=self,
#                 state=DatasetState.SUCCEED,
#                 details=mess,
#             )
#         self.Outputs.data.send(scan)

#     def reprocess(self, dataset):
#         self.process(dataset)


# class _OutputDir(qt.QGroupBox):
#     def __init__(self, parent):
#         qt.QGroupBox.__init__(self, parent)
#         self._button_grp = qt.QButtonGroup(self)
#         self._button_grp.setExclusive(True)
#         self.setLayout(qt.QGridLayout())
#         self._oneLevelCB = qt.QRadioButton(OutputType.ONE_LEVEL_UPPER.value, self)
#         self._button_grp.addButton(self._oneLevelCB)
#         self.layout().addWidget(self._oneLevelCB, 1, 0, 1, 3)
#         self._staticOpt = qt.QRadioButton("", self)
#         self._button_grp.addButton(self._staticOpt)
#         self.layout().addWidget(self._staticOpt, 2, 0, 1, 1)
#         self._qleFolder = qt.QLineEdit("", self)
#         self.layout().addWidget(self._qleFolder, 2, 1, 1, 1)
#         self._selectPB = qt.QPushButton("select", self)
#         self.layout().addWidget(self._selectPB, 2, 2, 1, 1)

#         # connect signal / slot
#         self._selectPB.released.connect(self._selectFolder)
#         self._qleFolder.editingFinished.connect(self._activeStaticOpt)

#         # set up
#         self._oneLevelCB.setChecked(True)

#     def _activeStaticOpt(self, *args, **kwargs):
#         self._staticOpt.setChecked(True)

#     def _selectFolder(self):  # pragma: no cover
#         dialog = qt.QFileDialog(self)
#         dialog.setWindowTitle("Select destination folder for symbolic link")
#         dialog.setModal(1)
#         dialog.setFileMode(qt.QFileDialog.DirectoryOnly)

#         if not dialog.exec_():
#             dialog.close()
#             return None

#         dest_folder = dialog.selectedFiles()[0]
#         self._qleFolder.setText(dest_folder)

#     def getOutputType(self) -> OutputType:
#         if self._oneLevelCB.isChecked():
#             return OutputType.ONE_LEVEL_UPPER
#         elif self._staticOpt.isChecked():
#             return OutputType.STATIC
#         else:
#             raise TypeError("Not managed")

#     def getOutputFolder(self):
#         return self._qleFolder.text()



import weakref
from tomwer.core.scan.scanbase import TomwerScanBase


class IcatProcessedDataBase:
    """
    Container class to get all the information about processed data to be publish
    """

    def __init__(self, data_dir: str, scan: TomwerScanBase) -> None:
        self._output_dir = data_dir
        # output dir for the screenshots. The default is expected to be provided by the task creating the screenshot.
        # it can be overwrite by the 'SaveScreenshotTask' if an output dir is provided
        self._scan = weakref.ref(scan) if scan is not None else scan

    @property
    def data_dir(self):
        return self._output_dir

    @property
    def scan(self):
        return self._scan() if self._scan is not None else None

    @property
    def metadata(self) -> dict:
        scan = self.scan
        if scan is not None:
            return scan.build_icat_metadata()
        else:
            return {}

    def to_dict(self):
        return {
            "output_dir": self.data_dir,
            "icat_metadata": self.metadata,
        }



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/04/2021"

from silx.gui import qt


class _IgnoreWheelBase:
    """QWidget giving priority to the scrollArea containing it"""

    def __init__(self, scrollArea):
        self._scrollArea = scrollArea

    def wheelEvent(self, *args, **kwargs):
        if self._scrollArea is not None:
            self._scrollArea.wheelEvent(*args, **kwargs)


class QComboBoxIgnoreWheel(qt.QComboBox, _IgnoreWheelBase):
    """Combobox which give priority to the scrollArea containing it"""

    def __init__(self, parent, scrollArea):
        qt.QComboBox.__init__(self, parent, scrollArea=scrollArea)
        _IgnoreWheelBase.__init__(self, scrollArea)

    def wheelEvent(self, *args, **kwargs):
        # Note: we are enforce to duplicate this due to an error on pyqt side.
        self._scrollArea.wheelEvent(*args, **kwargs)


class QDoubleSpinBoxIgnoreWheel(qt.QDoubleSpinBox):
    def __init__(self, parent, scrollArea):
        qt.QDoubleSpinBox.__init__(self, parent)
        _IgnoreWheelBase.__init__(self, scrollArea)

    def wheelEvent(self, *args, **kwargs):
        # Note: we are enforce to duplicate this due to an error on pyqt side.
        self._scrollArea.wheelEvent(*args, **kwargs)


class QSpinBoxIgnoreWheel(qt.QSpinBox, _IgnoreWheelBase):
    def __init__(self, parent, scrollArea):
        qt.QSpinBox.__init__(self, parent, scrollArea=scrollArea)
        _IgnoreWheelBase.__init__(self, scrollArea)

    def wheelEvent(self, *args, **kwargs):
        # Note: we are enforce to duplicate this due to an error on pyqt side.
        self._scrollArea.wheelEvent(*args, **kwargs)
